#
# Copyright 2014-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
from copy import deepcopy
import functools
import hashlib
import json
import logging
import os
import uuid

import yaml

from lago import log_utils, plugins, utils
from lago.config import config
from lago.providers.libvirt.network import BridgeNetwork, NATNetwork

LOGGER = logging.getLogger(__name__)
LogTask = functools.partial(log_utils.LogTask, logger=LOGGER)
log_task = functools.partial(log_utils.log_task, logger=LOGGER)


def _gen_ssh_command_id():
    return uuid.uuid1().hex[:8]


def _guestfs_copy_path(g, guest_path, host_path):
    if g.is_file(guest_path):
        with open(host_path, 'w') as f:
            f.write(g.read_file(guest_path))
    elif g.is_dir(guest_path):
        os.mkdir(host_path)
        for path in g.ls(guest_path):
            _guestfs_copy_path(
                g,
                os.path.join(
                    guest_path,
                    path,
                ),
                os.path.join(host_path, os.path.basename(path)),
            )


def _path_to_xml(basename):
    return os.path.join(
        os.path.dirname(__file__),
        basename,
    )


class VirtEnv(object):
    '''Env properties:
    * prefix
    * vms
    * net

    '''

    def __init__(self, prefix, vm_specs, net_specs):
        self.vm_types = plugins.load_plugins(
            plugins.PLUGIN_ENTRY_POINTS['vm'],
            instantiate=False,
        )
        self.prefix = prefix

        with open(self.prefix.paths.uuid(), 'r') as uuid_fd:
            self.uuid = uuid_fd.read().strip()

        self._nets = {}
        compat = self.get_compat()
        for name, spec in net_specs.items():
            self._nets[name] = self._create_net(spec, compat)

        self._vms = {}
        self._default_vm_type = config.get('default_vm_type')
        for name, spec in vm_specs.items():
            self._vms[name] = self._create_vm(spec)

    def _create_net(self, net_spec, compat):
        if net_spec['type'] == 'nat':
            cls = NATNetwork
        elif net_spec['type'] == 'bridge':
            cls = BridgeNetwork
        return cls(self, net_spec, compat=compat)

    def _create_vm(self, vm_spec):
        vm_type_name = vm_spec.get('vm-type', self._default_vm_type)
        try:
            vm_type = self.vm_types[vm_type_name]
        except KeyError:
            raise RuntimeError(
                'Unknown VM type: {0}, available types: {1}'.format(
                    vm_type_name, ','.join(self.vm_types.keys())
                )
            )
        vm_spec['vm-type'] = vm_type_name
        return vm_type(self, vm_spec)

    def prefixed_name(self, unprefixed_name, max_length=0):
        """
        Returns a uuid pefixed identifier

        Args:
            unprefixed_name(str): Name to add a prefix to
            max_length(int): maximum length of the resultant prefixed name,
                will adapt the given name and the length of the uuid ot fit it

        Returns:
            str: prefixed identifier for the given unprefixed name
        """
        if max_length == 0:
            prefixed_name = '%s-%s' % (self.uuid[:8], unprefixed_name)
        else:
            if max_length < 6:
                raise RuntimeError(
                    "Can't prefix with less than 6 chars (%s)" %
                    unprefixed_name
                )
            if max_length < 16:
                _uuid = self.uuid[:4]
            else:
                _uuid = self.uuid[:8]

            name_max_length = max_length - len(_uuid) - 1

            if name_max_length < len(unprefixed_name):
                hashed_name = hashlib.sha1(unprefixed_name).hexdigest()
                unprefixed_name = hashed_name[:name_max_length]

            prefixed_name = '%s-%s' % (_uuid, unprefixed_name)

        return prefixed_name

    def virt_path(self, *args):
        return self.prefix.paths.virt(*args)

    def bootstrap(self):
        vms = filter(
            lambda vm: vm.spec.get('bootstrap', True), self._vms.values()
        )
        if vms:
            utils.invoke_in_parallel(lambda vm: vm.bootstrap(), vms)

    def export_vms(
        self,
        vms_names,
        standalone,
        dst_dir,
        compress,
        init_file_name,
        out_format,
        collect_only=False,
        with_threads=True
    ):
        # todo: move this logic to PrefixExportManager
        if not vms_names:
            vms_names = self._vms.keys()

        running_vms = []
        vms = []
        for name in vms_names:
            try:
                vm = self._vms[name]
                if not vm.spec.get('skip-export'):
                    vms.append(vm)
                    if vm.running():
                        running_vms.append(vm)
            except KeyError:
                raise utils.LagoUserException(
                    'Entity {} does not exist'.format(name)
                )

        if running_vms:
            raise utils.LagoUserException(
                'The following vms must be off:\n{}'.format(
                    '\n'.join([_vm.name() for _vm in running_vms])
                )
            )

        with LogTask('Exporting disks to: {}'.format(dst_dir)):
            if not os.path.isdir(dst_dir):
                os.mkdir(dst_dir)

            def _export_disks(vm):
                return vm.export_disks(
                    standalone, dst_dir, compress, collect_only, with_threads
                )

            if collect_only:
                return (
                    reduce(
                        lambda x, y: x.update(y) or x, map(_export_disks, vms)
                    )
                )
            else:
                if with_threads:
                    results = utils.invoke_in_parallel(_export_disks, vms)
                else:
                    results = map(_export_disks, vms)

                results = reduce(lambda x, y: x.update(y) or x, results)

        self.generate_init(
            os.path.join(dst_dir, init_file_name), out_format, vms
        )

        results['init-file'] = os.path.join(dst_dir, init_file_name)

        return results

    def generate_init(self, dst, out_format, vms_to_include, filters=None):
        """
        Generate an init file which represents this env and can
        be used with the images created by self.export_vms

        Args:
            dst (str): path and name of the new init file
            out_format (plugins.output.OutFormatPlugin):
                formatter for the output (the default is yaml)
            filters (list): list of paths to keys that should be removed from
                the init file
            vms_to_include (list of :class:lago.plugins.vm.VMPlugin):
                list of vms to include in the init file
        Returns:
            None
        """
        # todo: move this logic to PrefixExportManager
        with LogTask('Exporting init file to: {}'.format(dst)):
            # Set the default formatter to yaml. The default formatter
            # doesn't generate a valid init file, so it's not reasonable
            # to use it
            if isinstance(out_format, plugins.output.DefaultOutFormatPlugin):
                out_format = plugins.output.YAMLOutFormatPlugin()

            if not filters:
                filters = [
                    'domains/*/disks/*/metadata',
                    'domains/*/metadata/deploy-scripts', 'domains/*/snapshots',
                    'domains/*/name', 'nets/*/mapping', 'nets/*/dns_records'
                ]

            spec = self.get_env_spec(filters)
            temp = {}

            for vm in vms_to_include:
                temp[vm.name()] = spec['domains'][vm.name()]

            spec['domains'] = temp

            for _, domain in spec['domains'].viewitems():
                domain['disks'] = [
                    d for d in domain['disks'] if not d.get('skip-export')
                ]

                for disk in domain['disks']:
                    if disk['type'] == 'template':
                        disk['template_type'] = 'qcow2'
                    elif disk['type'] == 'empty':
                        disk['type'] = 'file'
                        disk['make_a_copy'] = 'True'

                    # Insert the relative path to the exported images
                    disk['path'] = os.path.join(
                        '$LAGO_INITFILE_PATH', os.path.basename(disk['path'])
                    )

            with open(dst, 'wt') as f:
                if isinstance(out_format, plugins.output.YAMLOutFormatPlugin):
                    # Dump the yaml file without type tags
                    # TODO: Allow passing parameters to output plugins
                    f.write(yaml.safe_dump(spec))
                else:
                    f.write(out_format.format(spec))

    def get_env_spec(self, filters=None):
        """
        Get the spec of the current env.
        The spec will hold the info about all the domains and
        networks associated with this env.

        Args:
            filters (list): list of paths to keys that should be removed from
                the init file
        Returns:
            dict: the spec of the current env
        """
        spec = {
            'domains':
                {
                    vm_name: deepcopy(vm_object.spec)
                    for vm_name, vm_object in self._vms.viewitems()
                },
            'nets':
                {
                    net_name: deepcopy(net_object.spec)
                    for net_name, net_object in self._nets.viewitems()
                }
        }

        if filters:
            utils.filter_spec(spec, filters)

        return spec

    def start(self, vm_names=None):
        if not vm_names:
            log_msg = 'Start Prefix'
            vms = self._vms.values()
            nets = self._nets.values()
        else:
            log_msg = 'Start specified VMs'
            vms = [self._vms[vm_name] for vm_name in vm_names]
            nets = set()
            for vm in vms:
                nets = nets.union(
                    set(self._nets[net_name] for net_name in vm.nets())
                )

        with LogTask(log_msg), utils.RollbackContext() as rollback:
            with LogTask('Start nets'):
                for net in nets:
                    net.start()
                    rollback.prependDefer(net.stop)

            with LogTask('Start vms'):
                for vm in vms:
                    vm.start()
                    rollback.prependDefer(vm.stop)
                rollback.clear()

    def _get_stop_shutdown_common_args(self, vm_names):
        """
        Get the common arguments for stop and shutdown commands

        Args:
            vm_names (list of str): The names of the requested vms

        Returns
            list of plugins.vm.VMProviderPlugin:
                vms objects that should be stopped
            list of virt.Network: net objects that should be stopped
            str: log message

        Raises:
            utils.LagoUserException: If a vm name doesn't exist
        """

        vms_to_stop = self.get_vms(vm_names).values()

        if not vm_names:
            log_msg = '{} prefix'
            nets = self._nets.values()
        else:
            log_msg = '{} specified VMs'
            nets = self._get_unused_nets(vms_to_stop)

        return vms_to_stop, nets, log_msg

    def _get_unused_nets(self, vms_to_stop):
        """
        Return a list of nets that used only by the vms in vms_to_stop

        Args:
            vms_to_stop (list of str): The names of the requested vms

        Returns
            list of virt.Network: net objects that used only by
                vms in vms_to_stop

        Raises:
            utils.LagoUserException: If a vm name doesn't exist
        """

        vm_names = [vm.name() for vm in vms_to_stop]
        unused_nets = set()

        for vm in vms_to_stop:
            unused_nets = unused_nets.union(vm.nets())
        for vm in self._vms.values():
            if not vm.running() or vm.name() in vm_names:
                continue
            for net in vm.nets():
                unused_nets.discard(net)
        nets = [self._nets[net] for net in unused_nets]

        return nets

    def stop(self, vm_names=None):

        vms, nets, log_msg = self._get_stop_shutdown_common_args(vm_names)

        with LogTask(log_msg.format('Stop')):
            with LogTask('Stop vms'):
                for vm in vms:
                    vm.stop()
            with LogTask('Stop nets'):
                for net in nets:
                    net.stop()

    def shutdown(self, vm_names, reboot=False):

        vms, nets, log_msg = self._get_stop_shutdown_common_args(vm_names)

        if reboot:
            with LogTask(log_msg.format('Reboot')):
                with LogTask('Reboot vms'):
                    for vm in vms:
                        vm.reboot()
        else:
            with LogTask(log_msg.format('Shutdown')):
                with LogTask('Shutdown vms'):
                    for vm in vms:
                        vm.shutdown()
                with LogTask('Stop nets'):
                    for net in nets:
                        net.stop()

    def get_nets(self):
        return self._nets.copy()

    def get_net(self, name=None):
        if name:
            return self.get_nets().get(name)
        else:
            try:
                return [
                    net for net in self.get_nets().values()
                    if net.is_management()
                ].pop()
            except IndexError:
                return self.get_nets().values().pop()

    def get_vms(self, vm_names=None):
        """
        Returns the vm objects associated with vm_names
        if vm_names is None, return all the vms in the prefix

        Args:
            vm_names (list of str): The names of the requested vms

        Returns
            dict: Which contains the requested vm objects indexed by name

        Raises:
            utils.LagoUserException: If a vm name doesn't exist
        """
        if not vm_names:
            return self._vms.copy()

        missing_vms = []
        vms = {}
        for name in vm_names:
            try:
                vms[name] = self._vms[name]
            except KeyError:
                # TODO: add resolver by suffix
                missing_vms.append(name)

        if missing_vms:
            raise utils.LagoUserException(
                'The following vms do not exist: \n{}'.format(
                    '\n'.join(missing_vms)
                )
            )

        return vms

    def get_vm(self, name):
        return self._vms[name]

    @classmethod
    def from_prefix(cls, prefix):
        virt_path = functools.partial(prefix.paths.prefixed, 'virt')

        with open(virt_path('env'), 'r') as f:
            env_dom = json.load(f)

        net_specs = {}
        for name in env_dom['nets']:
            with open(virt_path('net-%s' % name), 'r') as f:
                net_specs[name] = json.load(f)

        vm_specs = {}
        for name in env_dom['vms']:
            with open(virt_path('vm-%s' % name), 'r') as f:
                vm_specs[name] = json.load(f)

        return cls(prefix, vm_specs, net_specs)

    @log_task('Save prefix')
    def save(self):
        with LogTask('Save nets'):
            for net in self._nets.values():
                net.save()

        with LogTask('Save VMs'):
            for vm in self._vms.values():
                vm.save()

        spec = {
            'nets': self._nets.keys(),
            'vms': self._vms.keys(),
        }

        with LogTask('Save env'):
            with open(self.virt_path('env'), 'w') as f:
                utils.json_dump(spec, f)

    @log_task('Create VMs snapshots')
    def create_snapshots(self, name):
        utils.invoke_in_parallel(
            lambda vm: vm.create_snapshot(name),
            self._vms.values(),
        )

    @log_task('Revert VMs snapshots')
    def revert_snapshots(self, name):
        utils.invoke_in_parallel(
            lambda vm: vm.revert_snapshot(name),
            self._vms.values(),
        )

    def get_snapshots(self, domains=None):
        """
        Get the list of snapshots for each domain

        Args:
            domanins(list of str): list of the domains to get the snapshots
            for, all will be returned if none or empty list passed

        Returns:
            dict of str -> list(str): with the domain names and the list of
            snapshots for each
        """
        snapshots = {}
        for vm_name, vm in self.get_vms().items():
            if domains and vm_name not in domains:
                continue

            snapshots[vm_name] = vm._spec['snapshots']

        return snapshots

    def get_compat(self):
        """Get compatibility level for this environment - which is the Lago
        version used to create this environment """
        # Prior to version 0.37.0, the version which the environment was
        # initialized in was not saved, so we default to 0.36.
        return self.prefix.metadata.get('lago_version', '0.36.0')



#
# Copyright 2016 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
"""
Service Plugin
=================

This plugins are used in order to manage services in the vms

"""
from abc import (abstractmethod, abstractproperty, ABCMeta)

from enum import Enum

from . import Plugin


class ServiceState(Enum):
    #: This state corresponds to a service that is not available in the domain
    MISSING = 0
    INACTIVE = 1
    ACTIVE = 2


class ServicePlugin(Plugin):
    __metaclass__ = ABCMeta

    def __init__(self, vm, name):
        self._vm = vm
        self._name = name

    @abstractmethod
    def state(self):
        """
        Check the current status of the service

        Returns:
            ServiceState: Which state the service is at right now
        """
        pass

    @abstractmethod
    def _request_start(self):
        """
        Low level implementation of the service start request, used by the
        `func:start` method

        Returns:
            bool: True if the service succeeded to start, False otherwise
        """
        pass

    @abstractmethod
    def _request_stop(self):
        """
        Low level implementation of the service stop request, used by the
        `func:stop` method

        Returns:
            bool: True if the service succeeded to stop, False otherwise
        """
        pass

    @abstractproperty
    def BIN_PATH(self):
        """
        Path to the binary used to manage services in the vm, will be checked
        for existence when trying to decide if the serviece is supported on the
        VM (see `func:is_supported`).

        Returns:
            str: Full path to the binary insithe the domain
        """
        pass

    def exists(self):
        return self.state() != ServiceState.MISSING

    def alive(self):
        return self.state() == ServiceState.ACTIVE

    def start(self):
        state = self.state()
        if state == ServiceState.MISSING:
            raise RuntimeError('Service %s not present' % self._name)
        elif state == ServiceState.ACTIVE:
            return

        if self._request_start():
            raise RuntimeError('Failed to start service')

    def stop(self):
        state = self.state()
        if state == ServiceState.MISSING:
            raise RuntimeError('Service %s not present' % self._name)
        elif state == ServiceState.INACTIVE:
            return

        if self._request_stop():
            raise RuntimeError('Failed to stop service')

    @classmethod
    def is_supported(cls, vm):
        return vm.ssh(['test', '-e', cls.BIN_PATH]).code == 0



from future.builtins import super
from abc import ABCMeta, abstractmethod
import logging
import functools
import copy
import os
from os import path
import time
import json
import shutil

from . import log_utils, utils

LOGGER = logging.getLogger(__name__)
LogTask = functools.partial(log_utils.LogTask, logger=LOGGER)


class DiskExportManager(object):
    """
    DiskExportManager object is responsible on the export process of
    an image from the current Lago prefix.

    DiskExportManger is the base class of specific DiskExportManger.
    Each specific DiskExportManger is responsible on the export process of an
    image with a specific type (e.g template, file...)

    Attributes:
        src (str): Path to the image that should be exported
        name (str): The name of the exported disk
        dst (str): The absolute path of the exported disk
        disk_type (str): The type of the image e.g template, file, empty...
        disk (dict): Disk attributes (of the disk that should be exported)
            as found in workdir/current/virt/VM-NAME
        exported_metadata(dict): A copy of the source disk metadata, this
            dict should be updated with new values during the export process.
        do_compress(bool): If true, apply compression to the exported
            disk.
    """

    __metaclass__ = ABCMeta

    def __init__(self, dst, disk_type, disk, do_compress):
        """
        Args:
            dst (str): The absolute path of the exported disk
            disk_type (str): The type of the image
                e.g template, file, empty...
            disk (dict): Disk attributes as found in
                workdir/current/virt/VM-NAME
            do_compress(bool): If true, apply compression to the
                exported disk.
        """
        self.src = path.expandvars(disk['path'])
        self.name = path.basename(self.src)
        self.dst = path.join(path.expandvars(dst), self.name)
        self.dist_type = disk_type
        self.disk = disk
        self.exported_metadata = copy.deepcopy(disk['metadata'])
        self.do_compress = do_compress

    @staticmethod
    def get_instance_by_type(dst, disk, do_compress, *args, **kwargs):
        """
        Args:
            dst (str): The path of the new exported disk.
                can contain env variables.
            disk (dict): Disk attributes
                (of the disk that should be exported) as
                found in workdir/current/virt/VM-NAME
            do_compress(bool): If true, apply compression to the exported
                disk.
        Returns
            An instance of a subclass of DiskExportManager which
            matches the disk type.
        """
        disk_type = disk['type']
        cls = HANDLERS.get(disk_type)
        if not cls:
            raise utils.LagoUserException(
                'Export is not supported for disk type {}'.format(disk_type)
            )
        return cls(dst, disk_type, disk, do_compress, *args, **kwargs)

    @abstractmethod
    def export(self):
        """
        This method will handle the export process and should
        implemented in each subclass.
        """
        raise NotImplementedError(
            'export is an abstract method that should'
            'be implemented in subclass'
        )

    def copy(self):
        """
        Copy the disk using cp in order to preserves the 'sparse'
        structure of the file
        """
        with LogTask('Copying disk'):
            utils.cp(self.src, self.dst)

    def sparse(self):
        """
        Make the exported images more compact by removing unused space.
        Please refer to 'virt-sparsify' for more info.
        """
        with LogTask('Making disk sparse'):
            # Removed unused space from the disk
            utils.sparse(self.dst, self.disk['format'])

    def calc_sha(self, checksum):
        """
        Calculate the checksum of the new exported disk, write it to
        a file, and update this managers 'exported_metadata'.

        Args:
            checksum(str): The type of the checksum
        """
        with LogTask('Calculating {}'.format(checksum)):
            with open(self.dst + '.hash', 'wt') as f:
                sha = utils.get_hash(self.dst, checksum)
                f.write(sha)
            self.exported_metadata[checksum] = sha

    def compress(self):
        """
        Compress the new exported image,
        Block size was taken from virt-builder page
        """
        if not self.do_compress:
            return
        with LogTask('Compressing disk'):
            utils.compress(self.dst, 16777216)
            os.unlink(self.dst)

    def update_lago_metadata(self):
        with LogTask('Updating Lago metadata'):
            self.exported_metadata['size'] = os.stat(self.dst).st_size
            self.exported_metadata['name'] = self.name
            self.exported_metadata['version'] = time.strftime("%Y%m%d.0")

    def write_lago_metadata(self):
        with LogTask('Writing Lago metadata'):
            with open(self.dst + '.metadata', 'wt') as f:
                json.dump(self.exported_metadata, f)


class TemplateExportManager(DiskExportManager):
    """
    TemplateExportManager is responsible exporting images of type template.

    Attributes:
        See superclass
    """

    def __init__(self, dst, disk_type, disk, do_compress, *args, **kwargs):
        super().__init__(dst, disk_type, disk, do_compress)
        self.standalone = kwargs['standalone']
        self.src_qemu_info = utils.get_qemu_info(self.src, backing_chain=True)

    def rebase(self):
        """
        Change the backing-file entry of the exported disk.
        Please refer to 'qemu-img rebase' manual for more info.
        """
        if self.standalone:
            rebase_msg = 'Merging layered image with base'
        else:
            rebase_msg = 'Rebase'

        with LogTask(rebase_msg):
            if len(self.src_qemu_info) == 1:
                # Base image (doesn't have predecessors)
                return

            if self.standalone:
                # Consolidate the layers and base image
                utils.qemu_rebase(target=self.dst, backing_file="")
            else:
                if len(self.src_qemu_info) > 2:
                    raise utils.LagoUserException(
                        'Layered export is currently supported for one '
                        'layer only.  You can try to use Standalone export.'
                    )
                # Put an identifier in the metadata of the copied layer,
                # this identifier will be used later by Lago in order
                # to resolve and download the base image
                parent = self.src_qemu_info[0]['backing-filename']

                # Hack for working with lago images naming convention
                # For example: /var/lib/lago/store/phx_repo:el7.3-base:v1
                # Extract only the image name and the version
                # (in the example el7.3-base:v1)
                parent = os.path.basename(parent)
                try:
                    parent = parent.split(':', 1)[1]
                except IndexError:
                    pass

                parent = './{}'.format(parent)
                utils.qemu_rebase(
                    target=self.dst, backing_file=parent, safe=False
                )

    def update_lago_metadata(self):
        super().update_lago_metadata()

        if self.standalone:
            self.exported_metadata['base'] = 'None'
        else:
            self.exported_metadata['base'] = os.path.basename(self.src)

    def export(self):
        """
           See DiskExportManager.export
        """
        with LogTask('Exporting disk {} to {}'.format(self.name, self.dst)):
            with utils.RollbackContext() as rollback:
                rollback.prependDefer(
                    shutil.rmtree, self.dst, ignore_errors=True
                )
                self.copy()
                self.sparse()
                self.rebase()
                self.calc_sha('sha1')
                self.update_lago_metadata()
                self.write_lago_metadata()
                self.compress()
                rollback.clear()


class FileExportManager(DiskExportManager):
    """
    FileExportManager is responsible exporting images of type file and empty.

    Attributes:
        standalone (bool): If true, create a new image which is the result of
            merging all the layers of src (the image that we want to export).
        src_qemu_info (dict): Metadata on src which was generated by qemu-img.
    """

    def __init__(self, dst, disk_type, disk, do_compress, *args, **kwargs):
        super().__init__(dst, disk_type, disk, do_compress)

    def export(self):
        """
            See DiskExportManager.export
        """
        with LogTask('Exporting disk {} to {}'.format(self.name, self.dst)):
            with utils.RollbackContext() as rollback:
                rollback.prependDefer(
                    shutil.rmtree, self.dst, ignore_errors=True
                )
                self.copy()
                if not self.disk['format'] == 'iso':
                    self.sparse()
                self.calc_sha('sha1')
                self.update_lago_metadata()
                self.write_lago_metadata()
                self.compress()
                rollback.clear()


class VMExportManager(object):
    """
    VMExportManager object is responsible on the export process of a list of
    disks.

    Attributes:
        disks (list of dicts): Disks to export.
        dst (str): Where to place the exported disks.
        compress(bool): If True compress each exported disk.
        with_threads(bool): If True, run the export in parallel
        *args(list): Extra args, will be passed to each
            DiskExportManager
        **kwargs(dict): Extra args, will be passed to each
            DiskExportManager

    """

    def __init__(
        self, disks, dst, compress, with_threads=True, *args, **kwargs
    ):
        self._disks = disks
        self._dst = os.path.expandvars(os.path.realpath(dst))
        self._compress = compress
        self._with_threads = with_threads
        self._args = args
        self._kwargs = kwargs

    def _collect(self):
        """
        Returns:
            (generator of dicts): The disks that needed to be exported
        """
        return (disk for disk in self._disks if not disk.get('skip-export'))

    def collect_paths(self):
        """
        Returns:
            (list of str): The path of the disks that will be exported.
        """
        return [os.path.expandvars(disk['path']) for disk in self._collect()]

    def exported_disks_paths(self):
        """
        Returns:
            (list of str): The path of the exported disks.
        """
        return [
            os.path.join(self._dst, os.path.basename(disk['path']))
            for disk in self._collect()
        ]

    def _get_export_mgr(self):
        """
        Returns:
            (DiskExportManager): Handler for each disk
        """
        return (
            DiskExportManager.get_instance_by_type(
                dst=self._dst,
                disk=disk,
                do_compress=self._compress,
                *self._args,
                **self._kwargs
            ) for disk in self._collect()
        )

    def export(self):
        """
        Run the export process
        Returns:
            (list of str): The path of the exported disks.
        """
        if self._with_threads:
            utils.invoke_different_funcs_in_parallel(
                *map(lambda mgr: mgr.export, self._get_export_mgr())
            )
        else:
            for mgr in self._get_export_mgr():
                mgr.export()

        return self.exported_disks_paths()


HANDLERS = {
    'file': FileExportManager,
    'empty': FileExportManager,
    'template': TemplateExportManager
}



import logging
import imp

LOGGER = logging.getLogger(__name__)


def check_import(module_name):
    """
    Search if a module exists, and it is possible to try importing it

    Args:
        module_name(str): module to import

    Returns:
        bool: True if the package is found
    """
    try:
        imp.find_module(module_name)
        return True
    except ImportError:
        return False



#
# Copyright 2014 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
import utils

_BRCTL = ['sudo', 'brctl']
_IP = ['sudo', 'ip']


def _brctl(command, *args):
    ret, out, err = utils.run_command(_BRCTL + [command] + list(args))
    if ret:
        raise RuntimeError(
            'brctl %s failed\nrc: %d\n\nout:\n%s\n\nerr:\n%s' %
            (command, ret, out, err)
        )
    return ret, out, err


def _set_link(name, state):
    ret, _, _ = utils.run_command(_IP + ['link', 'set', 'dev', name, state])
    if ret:
        raise RuntimeError('Could not set %s to state %s' % (name, state))


def create(name, stp=True):
    _brctl('addbr', name)
    try:
        _set_link(name, 'up')
        if stp:
            _brctl('stp', name, 'on')
    except:
        _brctl('delbr', name)
        raise


def destroy(name):
    _set_link(name, 'down')
    _brctl('delbr', name)


def exists(name):
    ret, out, _ = utils.run_command(
        ['ip', '-o', 'link', 'show', 'type', 'bridge']
    )
    if ret:
        raise RuntimeError('Failed to check if bridge {} exists'.format(name))

    for entry in out.splitlines():
        if name == entry.split(':')[1].strip():
            return True

    return False



#
# Copyright 2014-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
"""
Utilities to help deal with the libvirt python bindings
"""
import libvirt
import xmltodict
import lxml.etree
import logging
import pkg_resources
from jinja2 import Environment, PackageLoader, TemplateNotFound
from lago.config import config

LOGGER = logging.getLogger(__name__)

#: Mapping of domain statuses values to human readable strings
DOMAIN_STATES = {
    libvirt.VIR_DOMAIN_NOSTATE: 'no state',
    libvirt.VIR_DOMAIN_RUNNING: 'running',
    libvirt.VIR_DOMAIN_BLOCKED: 'blocked',  # the domain is blocked on resource
    libvirt.VIR_DOMAIN_PAUSED: 'paused',
    libvirt.VIR_DOMAIN_SHUTDOWN: 'begin shut down',
    libvirt.VIR_DOMAIN_SHUTOFF: 'shut off',
    libvirt.VIR_DOMAIN_CRASHED: 'crashed',
    libvirt.VIR_DOMAIN_PMSUSPENDED:
        'suspended',  # the domain is suspended by guest power management
}

DOMAIN_RUNNING_STATES = {
    libvirt.VIR_DOMAIN_RUNNING,
}

#: Singleton with the cached opened libvirt connections
LIBVIRT_CONNECTIONS = {}


class Domain(object):
    """
    Class to namespace libvirt domain related helpers
    """

    @staticmethod
    def resolve_state(state_number):
        """
        Get a nice description from a domain state number

        Args:
            state_number(list of int): State number as returned by
                :func:`libvirt.virDomain.state`

        Returns:
            str: small human readable description of the domain state, unknown
                if the state is not in the known list
        """
        return DOMAIN_STATES.get(state_number[0], 'unknown')


def libvirt_callback(userdata, err):
    pass


def auth_callback(credentials, user_data):
    for credential in credentials:
        if credential[0] == libvirt.VIR_CRED_AUTHNAME:
            credential[4] = config.get('libvirt_username')
        elif credential[0] == libvirt.VIR_CRED_PASSPHRASE:
            credential[4] = config.get('libvirt_password')

    return 0


def get_libvirt_connection(name):
    if name not in LIBVIRT_CONNECTIONS:
        auth = [
            [libvirt.VIR_CRED_AUTHNAME, libvirt.VIR_CRED_PASSPHRASE],
            auth_callback, None
        ]
        libvirt_url = config.get('libvirt_url', 'qemu:///system')

        libvirt.registerErrorHandler(f=libvirt_callback, ctx=None)
        LIBVIRT_CONNECTIONS[name] = libvirt.openAuth(libvirt_url, auth)

    return LIBVIRT_CONNECTIONS[name]


def get_template(basename):
    """
    Load a file as a string from the templates directory

    Args:
        basename(str): filename

    Returns:
        str: string representation of the file
    """
    return pkg_resources.resource_string(
        __name__, '/'.join(['templates', basename])
    )


def get_domain_template(distro, libvirt_ver, **kwargs):
    """
    Get a rendered Jinja2 domain template

    Args:
        distro(str): domain distro
        libvirt_ver(int): libvirt version
        kwargs(dict): args for template render

    Returns:
        str: rendered template
    """
    env = Environment(
        loader=PackageLoader('lago', 'providers/libvirt/templates'),
        trim_blocks=True,
        lstrip_blocks=True,
    )

    template_name = 'dom_template-{0}.xml.j2'.format(distro)
    try:
        template = env.get_template(template_name)
    except TemplateNotFound:
        LOGGER.debug('could not find template %s using default', template_name)
        template = env.get_template('dom_template-base.xml.j2')
    return template.render(libvirt_ver=libvirt_ver, **kwargs)


def dict_to_xml(spec, full_document=False):
    """
    Convert dict to XML

    Args:
        spec(dict): dict to convert
        full_document(bool): whether to add XML headers

    Returns:
        lxml.etree.Element: XML tree
    """

    middle = xmltodict.unparse(spec, full_document=full_document, pretty=True)
    return lxml.etree.fromstring(middle)



#!/usr/bin/env python2
#
# Copyright 2014-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
from __future__ import print_function

import argparse
import logging
import os
import pkg_resources
import sys
from textwrap import dedent
import warnings
from signal import signal, SIGTERM, SIGHUP

import lago
import lago.plugins
import lago.plugins.cli
import lago.templates
from lago.config import config
from lago import (log_utils, workdir as lago_workdir, utils, lago_ansible)
from lago.utils import (in_prefix, with_logging, LagoUserException)

LOGGER = logging.getLogger('cli')
in_lago_prefix = in_prefix(
    prefix_class=lago.prefix.Prefix,
    workdir_class=lago_workdir.Workdir,
)


@lago.plugins.cli.cli_plugin(
    help='Initialize a directory for framework deployment'
)
@lago.plugins.cli.cli_plugin_add_argument(
    'virt_config',
    help=(
        'Configuration of resources to deploy, json and yaml file formats '
        'are supported, takes option precedence over workdir. Will use '
        '$PWD/LagoInitFile by default. You can use any env vars in that file, '
        'inculuding the extra ones LAGO_PREFIX_PATH LAGO_WORKDIR_PATH and '
        'LAGO_INITFILE_PATH'
    ),
    metavar='VIRT_CONFIG',
    type=os.path.abspath,
    nargs='?',
)
@lago.plugins.cli.cli_plugin_add_argument(
    'workdir',
    help=(
        'Workdir directory of the deployment, if none passed, it will use '
        '$PWD/.lago'
    ),
    metavar='WORKDIR',
    type=os.path.abspath,
    nargs='?',
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--template-repo-path',
    help='Repo file describing the templates',
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--template-repo-name',
    help='Name of the repo from the template repos dir',
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--template-store',
    help='Location to store templates at',
    type=os.path.abspath,
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--template-repos',
    help='Location to store repos',
    type=os.path.abspath,
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--set-current',
    action='store_true',
    help='If passed, it will set the newly created prefix as the current one',
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--skip-bootstrap',
    action='store_true',
    help=(
        'If passed, will skip bootstrapping the images, useful if you are '
        'using templates and you already know they will have the correct '
        'root pass for example'
    ),
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--skip-build',
    action='store_true',
    help=(
        'If passed, will skip the build commands specified in the build'
        'section of all the disks'
    ),
)
def do_init(
    workdir,
    virt_config,
    prefix_name='default',
    template_repo_path=None,
    template_repo_name=None,
    template_store=None,
    template_repos=None,
    set_current=False,
    skip_bootstrap=False,
    skip_build=False,
    **kwargs
):

    if virt_config is None and workdir is not None:
        virt_config = workdir
        workdir = None

    if workdir is None:
        workdir = os.path.abspath('.lago')

    if virt_config is None:
        virt_config = os.path.abspath('LagoInitFile')
    if not os.path.isfile(virt_config):
        raise LagoUserException(
            'Unable to find init file: {0}'.format(virt_config)
        )

    os.environ['LAGO_INITFILE_PATH'] = os.path.dirname(
        os.path.abspath(virt_config)
    )

    if prefix_name == 'current':
        prefix_name = 'default'

    with log_utils.LogTask('Initialize and populate prefix', LOGGER):
        LOGGER.debug('Using workdir %s', workdir)
        workdir = lago_workdir.Workdir(workdir)
        if not (
            os.path.exists(workdir.path)
            and lago.workdir.Workdir.is_workdir(workdir.path)
        ):
            LOGGER.debug(
                'Initializing workdir %s with prefix %s',
                workdir.path,
                prefix_name,
            )
            prefix = workdir.initialize(prefix_name)
        else:
            LOGGER.debug(
                'Adding prefix %s to workdir %s',
                prefix_name,
                workdir.path,
            )
            prefix = workdir.add_prefix(prefix_name)

        log_utils.setup_prefix_logging(prefix.paths.logs())

        try:
            if template_repo_path:
                repo = lago.templates.TemplateRepository.from_url(
                    template_repo_path
                )
            elif template_repo_name:
                repo = lago.templates.find_repo_by_name(
                    name=template_repo_name
                )

            else:
                raise RuntimeError(
                    'No template repo was configured or specified'
                )

            store = lago.templates.TemplateStore(template_store)

            with open(virt_config, 'r') as virt_fd:
                prefix.virt_conf_from_stream(
                    virt_fd,
                    repo,
                    store,
                    do_bootstrap=not skip_bootstrap,
                    do_build=not skip_build,
                )

            if set_current:
                workdir.set_current(new_current=prefix_name)

        except:
            workdir.cleanup()
            raise

        return workdir, prefix


@lago.plugins.cli.cli_plugin(help='Clean up deployed resources')
@in_lago_prefix
@with_logging
def do_cleanup(prefix, **kwargs):
    prefix.cleanup()


@lago.plugins.cli.cli_plugin(
    help='Cleanup and remove the whole prefix and any files in it'
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--all-prefixes',
    help="Destroy all the prefixes in the workdir",
    action='store_true',
)
@lago.plugins.cli.cli_plugin_add_argument(
    '-y',
    '--yes',
    help="Don't ask for confirmation, assume yes",
    action='store_true',
)
@in_lago_prefix
@with_logging
def do_destroy(
    prefix, yes, all_prefixes, parent_workdir, prefix_name, **kwargs
):

    if all_prefixes:
        warn_message = 'all the prefixes under ' + parent_workdir.path
        path = parent_workdir.path
    else:
        warn_message = prefix.paths.prefix_path()
        path = warn_message

    if not yes:
        response = raw_input(
            'Do you really want to destroy %s? [Yn] ' % warn_message
        )
        if response and response[0] not in 'Yy':
            LOGGER.info('Aborting on user input')
            return

    if os.path.islink(path):
        os.unlink(path)
        return

    if all_prefixes:
        parent_workdir.destroy()
    elif parent_workdir:
        parent_workdir.destroy([prefix_name])
    else:
        prefix.destroy()


@lago.plugins.cli.cli_plugin(help='Deploy lago resources')
@lago.plugins.cli.cli_plugin_add_argument(
    'vm_names',
    help='Name of the vm to start',
    metavar='VM_NAME',
    nargs='*',
)
@in_lago_prefix
@with_logging
def do_start(prefix, vm_names=None, **kwargs):
    prefix.start(vm_names=vm_names)


@lago.plugins.cli.cli_plugin(help='Destroy lago resources')
@lago.plugins.cli.cli_plugin_add_argument(
    'vm_names',
    help='Name of the vm to stop',
    metavar='VM_NAME',
    nargs='*',
)
@in_lago_prefix
@with_logging
def do_stop(prefix, vm_names, **kwargs):
    prefix.stop(vm_names=vm_names)


@lago.plugins.cli.cli_plugin(
    help='Shutdown and destroy, or reboot vms',
    description='This command will shutdown or reboot the given vms. '
    'Networks that will not have running vms connected to them after '
    'running this command will be stopped as well.'
)
@lago.plugins.cli.cli_plugin_add_argument(
    '-r',
    '--reboot',
    help='If specified, reboot the requested vms',
    action='store_true',
)
@lago.plugins.cli.cli_plugin_add_argument(
    'vm_names',
    help='Name of the vms to shutdown',
    metavar='VM_NAME',
    nargs='*',
)
@in_lago_prefix
@with_logging
def do_shutdown(prefix, vm_names, reboot, **kwargs):
    prefix.shutdown(vm_names, reboot)


@lago.plugins.cli.cli_plugin(
    help='Export virtual machine disks',
    description='This command will export the disks of the given vms. '
    'The disks of the vms will be exported to the '
    'current directory or to the path that was specified with "-d". '
    'If "-s" was specified, the disks will be exported as '
    '"standalone disk", which means that they will be merged with their '
    'base images. This command does not modifying the source disk. '
    'The env should be in "down" state when running this command.'
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--vm-names',
    '-n',
    help='Name of the vms to export. If no name is specified, export all '
    'the vms in this prefix.',
    metavar='VM_NAME',
    nargs='*',
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--standalone',
    '-s',
    help='If not specified, export a layered image',
    action='store_true',
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--compress',
    '-c',
    help='If specified, compress the exported images with xz',
    action='store_true',
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--dst-dir',
    '-d',
    default='.',
    help='Dir to place the exported images in',
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--init-file-name',
    default='LagoInitFile',
    help='The name of the exported init file',
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--collect-only',
    action='store_true',
    help='Only output the disks that will be exported',
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--without-threads',
    action='store_true',
    help='If set, do not use threads',
)
@in_lago_prefix
@with_logging
def do_export(
    prefix, vm_names, standalone, dst_dir, compress, init_file_name,
    out_format, collect_only, without_threads, **kwargs
):
    output = prefix.export_vms(
        vm_names, standalone, dst_dir, compress, init_file_name, out_format,
        collect_only, not without_threads
    )
    if collect_only:
        print(out_format.format(output))


@lago.plugins.cli.cli_plugin(
    help='Create snapshots for all deployed resources'
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--list',
    '-l',
    dest='list_only',
    help='List current available snapshots',
    action='store_true',
)
@lago.plugins.cli.cli_plugin_add_argument(
    'snapshot_name',
    help='Name of the snapshot to create',
    metavar='SNAPSHOT_NAME',
    nargs='?',
    default=None,
)
@in_lago_prefix
@with_logging
def do_snapshot(prefix, list_only, snapshot_name, out_format, **kwargs):
    if list_only:
        snapshots = prefix.get_snapshots()
        print(out_format.format(snapshots))
    elif snapshot_name:
        prefix.create_snapshots(snapshot_name)
    else:
        raise RuntimeError('No snapshot name provided')


@lago.plugins.cli.cli_plugin(help='Revert resources to a snapshot')
@lago.plugins.cli.cli_plugin_add_argument(
    'snapshot_name',
    help='Name of the snapshot to revert to',
    metavar='SNAPSHOT_NAME',
)
@in_lago_prefix
@with_logging
def do_revert(prefix, snapshot_name, **kwargs):
    prefix.revert_snapshots(snapshot_name)


@lago.plugins.cli.cli_plugin(
    help='Open shell on the domain or run as script/command',
    prefix_chars='\x00',
)
@lago.plugins.cli.cli_plugin_add_argument(
    'args',
    help=(
        'If none provided, an interactive shell will be started.\n'
        'If arguments start with -c, what follows will be '
        'executes as a command.\n'
        'Otherwise, if a single provided, it will be ran as script'
        ' on the domain.'
    ),
    nargs='*',
)
@lago.plugins.cli.cli_plugin_add_argument(
    'host',
    help='Host to connect to',
    metavar='HOST',
)
@in_lago_prefix
@with_logging
def do_shell(prefix, host, args=None, **kwargs):
    args = args or []
    try:
        host = prefix.virt_env.get_vm(host)
    except KeyError:
        ssh_host = None
        for possible_host in prefix.virt_env.get_vms():
            if possible_host.endswith(host):
                ssh_host = prefix.virt_env.get_vm(possible_host)
                break
        if ssh_host:
            host = ssh_host
        else:
            LOGGER.error('Unable to find VM %s', host)
            LOGGER.info(
                'Available VMs:\n\t' +
                '\n\t'.join(prefix.virt_env.get_vms().keys())
            )
            raise

    if not host.running():
        raise RuntimeError(
            'Host %s is not "running", but "%s"' % (host.name(), host.state())
        )

    host.wait_for_ssh()

    if len(args) == 0:
        result = host.interactive_ssh(['bash'])
    elif len(args) == 1 and os.path.isfile(args[0]):
        result = host.ssh_script(args[0])
    else:
        if args[0] == '-c':
            args = args[1:]

        result = host.interactive_ssh(args)

    sys.exit(result.code)


@lago.plugins.cli.cli_plugin(
    help='Open serial console to the domain',
)
@lago.plugins.cli.cli_plugin_add_argument(
    'host',
    help='Host to connect to',
    metavar='HOST',
)
@in_lago_prefix
@with_logging
def do_console(prefix, host, **kwargs):
    try:
        host = prefix.virt_env.get_vm(host)
    except KeyError:
        LOGGER.error('Unable to find VM %s', host)
        LOGGER.info(
            'Available VMs:\n\t' +
            '\n\t'.join(prefix.virt_env.get_vms().keys())
        )
        raise

    result = host.interactive_console()
    sys.exit(result.code)


@lago.plugins.cli.cli_plugin(
    help='Create Ansible host inventory of the environment',
    description=dedent(
        """
    This method iterates through all the VMs and creates an Ansible
    host inventory. For each vm it defines an IP address and a private key.

    The default groups are based on the values which associated
    with the following keys: 'vm-type', 'groups', 'vm-provider'.

    The 'keys' parameter can be used to override the default groups,
    for example to create a group which based on a 'service_provider',
    --keys 'service_provider' should be added to this command.

    Nested keys can be used also by specifying the path to the key,
    for example '/disks/0/metadata/distro' will create a group based on the
    distro of the os installed on disk at position 0 in the init file.
    (we assume that numeric values in the path should be used as index for
    list access).

    If the value associated with the key is a list, a group will be created
    for every item in that list (this is useful when you want to associate
    a machine with more than one group).

    The output of this command is printed to standard output.

    Example of a possible output:

    lago ansible_hosts -k 'vm-type'

    [vm-type=ovirt-host]
    lago-host1 ansible_host=1.2.3.4 ansible_ssh_private_key_file=/path/rsa
    lago-host2 ansible_host=1.2.3.6 ansible_ssh_private_key_file=/path/rsa
    [vm-type=ovirt-engine]
    lago-engine ansible_host=1.2.3.5 ansible_ssh_private_key_file=/path/rsa

    lago ansible_hosts -k 'disks/0/metadata/arch' 'groups'

    [disks/0/metadata/arch=x86_64]
    vm0-server ansible_host=1.2.3.4 ansible_ssh_private_key_file=/path/rsa
    vm1-slave ansible_host=1.2.3.5 ansible_ssh_private_key_file=/path/rsa
    vm2-slave ansible_host=1.2.3.6 ansible_ssh_private_key_file=/path/rsa
    [groups=slaves]
    vm1-slave ansible_host=1.2.3.5 ansible_ssh_private_key_file=/path/rsa
    vm2-slave ansible_host=1.2.3.6 ansible_ssh_private_key_file=/path/rsa
    [groups=servers]
    vm0-server ansible_host=1.2.3.4 ansible_ssh_private_key_file=/path/rsa
    """
    )
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--keys',
    '-k',
    help='Path to the keys that should be used as groups',
    default=['vm-type', 'groups', 'vm-provider'],
    metavar='KEY',
    nargs='*',
)
@in_lago_prefix
@with_logging
def do_generate_ansible_hosts(prefix, keys, **kwargs):
    print(lago_ansible.LagoAnsible(prefix).get_inventory_str(keys))


@lago.plugins.cli.cli_plugin(
    help='Show status of the deployed virtual resources'
)
@in_lago_prefix
@with_logging
def do_status(prefix, out_format, **kwargs):

    with open(prefix.paths.uuid()) as f:
        uuid = f.read()

    info_dict = {
        'Prefix':
            {
                'Base directory':
                    prefix.paths.prefix_path(),
                'UUID':
                    uuid,
                'Networks':
                    dict(
                        (
                            net.name(), {
                                'gateway': net.gw(),
                                'status': net.alive() and 'up' or 'down',
                                'management': net.is_management(),
                            }
                        ) for net in prefix.virt_env.get_nets().values()
                    ),
                'VMs':
                    dict(
                        (
                            vm.name(), {
                                'distro':
                                    vm.distro(),
                                'root password':
                                    vm.root_password(),
                                'status':
                                    vm.state(),
                                'snapshots':
                                    ', '.join(vm._spec['snapshots'].keys()),
                                'metadata':
                                    vm.metadata,
                                'NICs':
                                    dict(
                                        (
                                            'eth%d' % i, {
                                                'network': nic['net'],
                                                'ip': nic.get('ip', 'N/A'),
                                            }
                                        ) for i, nic in enumerate(vm.nics())
                                    ),
                            }
                        ) for vm in prefix.virt_env.get_vms().values()
                    ),
            },
    }

    print(out_format.format(info_dict))


@lago.plugins.cli.cli_plugin(help='List the prefixes (envs) in a Workdir')
@lago.plugins.cli.cli_plugin_add_argument(
    'workdir_path',
    help=dedent(
        """
        Path to the Workdir. If not provided Lago will
        try to find a Workdir relative to the current directory.
        """
    ),
    metavar='WORKDIR_PATH',
    type=os.path.abspath,
    nargs='?',
)
def do_list(workdir_path, out_format, **kwargs):
    if not workdir_path:
        workdir_path = lago_workdir.Workdir.resolve_workdir_path()

    workdir = lago_workdir.Workdir(path=workdir_path)
    workdir.load()
    resources = workdir.prefixes.keys()

    print(out_format.format(resources))


@lago.plugins.cli.cli_plugin(
    help='Copy file from a virtual machine to local machine'
)
@lago.plugins.cli.cli_plugin_add_argument(
    'local_path',
    help='Path on the local host to copy the file/dir to',
    metavar='LOCAL_PATH',
)
@lago.plugins.cli.cli_plugin_add_argument(
    'remote_path',
    help='Path of the file/dir to copy from the host',
    metavar='REMOTE_PATH',
)
@lago.plugins.cli.cli_plugin_add_argument(
    'host',
    help='Host to copy files from',
    metavar='HOST',
)
@in_lago_prefix
@with_logging
def do_copy_from_vm(prefix, host, remote_path, local_path, **kwargs):
    try:
        host = prefix.virt_env.get_vm(host)
    except KeyError:
        LOGGER.error('Unable to find VM %s', host)
        LOGGER.info(
            'Available VMs:\n\t' +
            '\n\t'.join(prefix.virt_env.get_vms().keys())
        )
        raise

    if not host.running():
        raise RuntimeError(
            'Host %s is not "running", but "%s"' % (host.name(), host.state())
        )

    host.wait_for_ssh()
    host.copy_from(remote_path, local_path)


@lago.plugins.cli.cli_plugin(
    help='Copy file/dir to a virtual machine from the local host'
)
@lago.plugins.cli.cli_plugin_add_argument(
    'remote_path',
    help='Local path to copy the file/dir to',
    metavar='REMOTE_PATH',
)
@lago.plugins.cli.cli_plugin_add_argument(
    'local_path',
    help='Path of the file/dir to copy from the host',
    metavar='LOCAL_PATH',
)
@lago.plugins.cli.cli_plugin_add_argument(
    'host',
    help='Host to copy files to',
    metavar='HOST',
)
@in_lago_prefix
@with_logging
def do_copy_to_vm(prefix, host, remote_path, local_path, **kwargs):
    try:
        host = prefix.virt_env.get_vm(host)
    except KeyError:
        LOGGER.error('Unable to find VM %s', host)
        LOGGER.info(
            'Available VMs:\n\t' +
            '\n\t'.join(prefix.virt_env.get_vms().keys())
        )
        raise

    if not host.running():
        raise RuntimeError(
            'Host %s is not "running", but "%s"' % (host.name(), host.state())
        )

    host.wait_for_ssh()
    host.copy_to(local_path, remote_path)


@lago.plugins.cli.cli_plugin(
    help=(
        'Collect logs from VMs, list of collected logs '
        'can be specified in the init file, under '
        'artifacts parameter '
    )
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--output',
    help='Path to place all the extracted at',
    required=True,
    type=os.path.abspath,
)
@lago.plugins.cli.cli_plugin_add_argument(
    '--no-skip',
    help='do not skip missing paths',
    action='store_true',
)
@in_lago_prefix
@with_logging
def do_collect(prefix, output, no_skip, **kwargs):
    prefix.collect_artifacts(output, ignore_nopath=not no_skip)


@lago.plugins.cli.cli_plugin(
    help='Run scripts that install necessary RPMs and configuration'
)
@in_lago_prefix
@with_logging
def do_deploy(prefix, **kwargs):
    prefix.deploy()


@lago.plugins.cli.cli_plugin(help="Dump configuration file")
@lago.plugins.cli.cli_plugin_add_argument(
    '--verbose',
    help='Include parameters with no default value.',
    action='store_true',
    default=False,
)
def do_generate(verbose, **kwargs):
    print(config.get_ini(incl_unset=verbose))


def create_parser(cli_plugins, out_plugins):
    parser = argparse.ArgumentParser(
        description='Command line interface to Lago',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '-l',
        '--loglevel',
        choices=['info', 'debug', 'error', 'warning'],
        help='Log level to use'
    )
    parser.add_argument(
        '--logdepth', type=int, help='How many task levels to show'
    )

    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s ' + pkg_resources.get_distribution("lago").version,
    )
    parser.add_argument(
        '--out-format',
        '-f',
        action='store',
        default='default',
        choices=out_plugins.keys(),
    )
    parser.add_argument(
        '--prefix-path',
        '-p',
        action='store',
        default=None,
        help=(
            'Path to the prefix to use, will be deprecated, use '
            '--workdir-path instead'
        ),
    )
    parser.add_argument(
        '--workdir-path',
        '-w',
        action='store',
        default=None,
        help='Path to the workdir to use.',
    )
    parser.add_argument(
        '--prefix-name',
        '-P',
        action='store',
        help='Name of the prefix to use.',
    )
    parser.add_argument(
        '--ssh-user',
        action='store',
        help='User for SSH provider.',
    )
    parser.add_argument(
        '--ssh-password',
        action='store',
        help='Password for SSH provider.',
    )
    parser.add_argument(
        '--ssh-tries',
        action='store',
        type=int,
        help='Number of ssh time outs to wait before failing.',
    )
    parser.add_argument(
        '--ssh-timeout',
        action='store',
        type=int,
        help='Seconds to wait before marking SSH connection as failed.'
    )
    parser.add_argument(
        '--libvirt_url',
        action='store',
        help='libvirt URI, currently only '
        'system'
        ' is supported.'
    )
    parser.add_argument(
        '--libvirt-user',
        action='store',
        help='libvirt user',
    )
    parser.add_argument(
        '--libvirt-password',
        action='store',
        help='libvirt password',
    )
    parser.add_argument(
        '--default_vm_type',
        action='store',
        help='Default vm type',
    )

    parser.add_argument(
        '--default_vm_provider',
        action='store',
        help='Default vm provider',
    )
    parser.add_argument(
        '--default_root_password',
        action='store',
        help='Default root password',
    )
    parser.add_argument(
        '--lease_dir',
        action='store',
        help='Path to store created subnets configurations'
    )
    parser.add_argument(
        '--reposync-dir',
        action='store',
        help='Reposync dir if used',
    )

    parser.add_argument('--ignore-warnings', action='store_true')
    parser.set_defaults(**config.get_section('lago', {}))
    verbs_parser = parser.add_subparsers(dest='verb', metavar='VERB')
    for cli_plugin_name, cli_plugin in cli_plugins.items():
        plugin_parser = verbs_parser.add_parser(
            name=cli_plugin_name,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            **cli_plugin.init_args
        )
        cli_plugin.populate_parser(plugin_parser)
        plugin_parser.set_defaults(**config.get_section(cli_plugin_name, {}))
    config.update_parser(parser=parser)

    return parser


def exit_handler(signum, frame):
    """
    Catch SIGTERM and SIGHUP and call "sys.exit" which raises
    "SystemExit" exception.
    This will trigger all the cleanup code defined in ContextManagers
    and "finally" statements.

    For more details about the arguments see "signal" documentation.

    Args:
        signum(int): The signal's number
        frame(frame): The current stack frame, can be None
    """

    LOGGER.debug('signal {} was caught'.format(signum))
    sys.exit(128 + signum)


def main():

    # Trigger cleanup on SIGTERM and SIGHUP
    signal(SIGTERM, exit_handler)
    signal(SIGHUP, exit_handler)

    cli_plugins = lago.plugins.load_plugins(
        lago.plugins.PLUGIN_ENTRY_POINTS['cli']
    )
    out_plugins = lago.plugins.load_plugins(
        lago.plugins.PLUGIN_ENTRY_POINTS['out']
    )
    parser = create_parser(
        cli_plugins=cli_plugins,
        out_plugins=out_plugins,
    )
    args = parser.parse_args()
    config.update_args(args)

    logging.basicConfig(level=logging.DEBUG)
    logging.root.handlers = [
        log_utils.TaskHandler(
            task_tree_depth=args.logdepth,
            level=getattr(logging, args.loglevel.upper()),
            dump_level=logging.ERROR,
            formatter=log_utils.ColorFormatter(fmt='%(msg)s', )
        )
    ]

    logging.captureWarnings(True)
    if args.ignore_warnings:
        logging.getLogger('py.warnings').setLevel(logging.ERROR)
    else:
        warnings.formatwarning = lambda message, *args, **kwargs: message

    args.out_format = out_plugins[args.out_format]
    if args.prefix_path:
        warnings.warn(
            'The option --prefix-path is going to be deprecated, use '
            '--workdir and --prefix instead',
            DeprecationWarning,
        )

    try:
        cli_plugins[args.verb].do_run(args)
    except utils.LagoException as e:
        LOGGER.info(e.message)
        LOGGER.debug(e)
        sys.exit(2)
    except Exception:
        LOGGER.exception('Error occured, aborting')
        sys.exit(1)


if __name__ == '__main__':
    main()



import logging
import functools
from lago import log_utils, utils
from collections import namedtuple

LOGGER = logging.getLogger(__name__)
LogTask = functools.partial(log_utils.LogTask, logger=LOGGER)
Command = namedtuple('Command', ['name', 'cmd'])


class Build(object):
    """
    A Build object represents a build section in the init file.
    Each build section (which in turn belongs to a specific disk)
    should get his own Build object

    In order to add support for a new build command, a new function with
    the name of the command should be implemented in this class. this function
    should accept a list of options and arguments and return a named tuple
    'Command', where 'Command.name' is the name of the command and
    'Command.cmd' is the a list containing the command and its args,
    for example:
    Command.name = 'virt-customize'
    Command.cmd = ['virt-customize', '-a', PATH_TO_DISK, SOME_CMDS...]

    Attributes:
        name (str): The name of the vm this builder belongs
        disk_path (str): The path to the disk that needs to be customized
        paths (lago.paths.Paths): The paths of the current prefix
        build_cmds (list of str): A list of commands that should
            be invoked on the disk
            located in disk_path
    """

    @staticmethod
    def normalize_options(options):
        """
        Turns a mapping of 'option: arg' to a list and prefix the options.
        arg can be a list of arguments.

        for example:

        dict = {
            o1: a1,
            o2: ,
            o3: [a31, a32]
            o4: []
        }

        will be transformed to:

        [
            prefix_option(o1), a1, prefix_option(o2),
            prefix_option(o3), a31, prefix_option(o3), a32
            prefix_option(o4)
        ]

        note that empty arguments are omitted

        Args:
            options (dict): A mapping between options and arguments

        Returns:
            lst: A normalized version of 'options' as mentioned above
        """
        normalized_options = []

        def _add(option, arg=None):
            normalized_options.append(option)
            arg and normalized_options.append(arg)

        for option, arg in options.viewitems():
            prefixed_option = Build.prefix_option(option)
            if isinstance(arg, list) and arg:
                for a in arg:
                    _add(prefixed_option, a)
            else:
                _add(prefixed_option, arg)

        return normalized_options

    @staticmethod
    def prefix_option(option):
        """
        Depends on the option's length, prefix it with '-' or '--'
        Args:
            option (str): The option to prefix
        Returns:
            str: prefixed option
        """
        if len(option) == 1:
            return '-{}'.format(option)
        else:
            return '--{}'.format(option)

    @classmethod
    def get_instance_from_build_spec(cls, name, disk_path, build_spec, paths):
        """
        Args:
            name (str): The name of the vm this builder belongs
            disk_path (str): The path to the disk that needs to be customized
            paths (lago.paths.Paths): The paths of the current prefix
            build_spec (dict): The build spec part, associated with the
                disk located at disk_path, from the init file.

        Returns:
            An instance of Build with a normalized build spec i.e ready to
                be invoked.
        """
        instance = cls(name, disk_path, paths)
        instance.normalize_build_spec(build_spec)
        return instance

    def __init__(self, name, disk_path, paths):
        self.name = name
        self.disk_path = disk_path
        self.paths = paths
        self.build_cmds = []

    def normalize_build_spec(self, build_spec):
        """
        Convert a build spec into a list of Command tuples.
        After running this command, self.build_cmds should hold all
        the commands that should be run on the disk in self.disk_path.

        Args:
            build_spec (dict): The buildspec part from the init file
        """
        for cmd in build_spec:
            if not cmd:
                continue
            cmd_name = cmd.keys()[0]
            cmd_options = cmd.values()[0]
            cmd_handler = self.get_cmd_handler(cmd_name)
            self.build_cmds.append(cmd_handler(cmd_options))

    def get_cmd_handler(self, cmd):
        """
        Return an handler for cmd.
        The handler and the command should have the same name.
        See class description for more info about handlers.

        Args:
            cmd (str): The name of the command

        Returns:
            callable: which handles cmd

        Raises:
            lago.build.BuildException: If an handler for cmd doesn't exist
        """
        cmd = cmd.replace('-', '_')
        handler = getattr(self, cmd, None)
        if not handler:
            raise BuildException(
                'Command {} is not supported as a '
                'build command'.format(cmd)
            )
        return handler

    def virt_customize(self, options):
        """
        Handler for 'virt-customize'
        note: if 'ssh-inject' option was specified without a path to a key,
        the prefix' key will be copied to the vm.

        Args:
            options (lst of str): Options and arguments for 'virt-customize'

        Returns:
            callable: which handles cmd

        Raises:
            lago.build.BuildException: If an handler for cmd doesn't exist
        """
        cmd = ['virt-customize', '-a', self.disk_path]
        if 'ssh-inject' in options and not options['ssh-inject']:
            options['ssh-inject'] = 'root:file:{}'.format(
                self.paths.ssh_id_rsa_pub()
            )

        options = self.normalize_options(options)
        cmd.extend(options)
        return Command('virt-customize', cmd)

    def build(self):
        """
        Run all the commands in self.build_cmds

        Raises:
            lago.build.BuildException: If a command returned a non-zero code
        """
        if not self.build_cmds:
            LOGGER.debug('No build commands were found, skipping build step')

        with LogTask('Building {} disk {}'.format(self.name, self.disk_path)):
            for command in self.build_cmds:
                with LogTask('Running command {}'.format(command.name)):
                    LOGGER.debug(command.cmd)
                    result = utils.run_command(command.cmd)
                    if result:
                        raise BuildException(result.err)


class BuildException(utils.LagoException):
    pass



#
# Copyright 2014 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
import os


class Paths(object):
    """
    A Paths object contains methods for getting the paths to the
    directories and files which composes the prefix.

    Attributes:
        _prefix_path (str): Path to the directory of the prefix
    """

    def __init__(self, prefix_path):
        """
        Args:
            prefix_path (str): Path to the directory of the prefix
        """
        # self._prefix should be dropped in lago ver 0.44
        self.prefix = prefix_path
        self._prefix_path = prefix_path

    def prefixed(self, *args):
        return os.path.join(self._prefix_path, *args)

    def prefix_path(self):
        return self._prefix_path

    def uuid(self):
        return self.prefixed('uuid')

    def ssh_id_rsa(self):
        return self.prefixed('id_rsa')

    def ssh_id_rsa_pub(self):
        return self.prefixed('id_rsa.pub')

    def images(self, *path):
        return self.prefixed('images', *path)

    def virt(self, *path):
        return self.prefixed('virt', *path)

    def logs(self):
        return self.prefixed('logs')

    def metadata(self):
        return self.prefixed('metadata')

    def prefix_lagofile(self):
        "This file represents a prefix that's initialized"
        return self.prefixed('initialized')

    def scripts(self, *args):
        return self.prefixed('scripts', *args)



#
# Copyright 2014 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
"""
About OutFormatPlugins

An OutFormatPlugin is used to format the output of the commands that extract
information from the perfixes, like status.
"""

import collections
import json
import yaml
from abc import (abstractmethod, ABCMeta)
import copy
from operator import itemgetter

from . import Plugin


class OutFormatPlugin(Plugin):
    __metaclass__ = ABCMeta

    def __init__(self):
        pass

    @abstractmethod
    def format(self, info_dict):
        """
        Execute any actions given the arguments

        Args:
            info_dict (dict): information to reformat

        Returns:
            str: String representing the formatted info
        """
        pass


class DefaultOutFormatPlugin(OutFormatPlugin):
    indent_unit = '    '

    def format(self, info_obj, indent=''):
        formatted_lines = []
        if isinstance(info_obj, list):
            if indent:
                formatted_lines.append('')
            for elem in info_obj:
                value_str = self.format(elem)
                formatted_lines.append(indent + value_str)
        elif isinstance(info_obj, collections.Mapping):
            for key in sorted(info_obj.keys()):
                value = info_obj[key]
                if isinstance(value, collections.Mapping):
                    if not value:
                        continue

                    formatted_lines.append('%s[%s]:' % (indent, str(key)))
                    value_str = self.format(
                        info_obj=value,
                        indent=indent + self.indent_unit,
                    )
                    if value_str:
                        formatted_lines.append(value_str)

                elif value not in (None, ''):
                    formatted_lines.append(
                        indent + str(key) + ': ' + self.format(
                            info_obj=value,
                            indent=indent + self.indent_unit,
                        )
                    )
        else:
            formatted_lines.append(str(info_obj))

        return '\n'.join(formatted_lines)


class JSONOutFormatPlugin(OutFormatPlugin):
    def format(self, info_dict):
        return json.dumps(
            info_dict,
            sort_keys=True,
            indent=4,
        )


class YAMLOutFormatPlugin(OutFormatPlugin):
    def format(self, info_dict):
        return yaml.dump(info_dict)


class FlatOutFormatPlugin(OutFormatPlugin):
    def format(self, info_dict, delimiter='/'):
        """
        This formatter will take a data structure that
        represent a tree and will print all the paths
        from the root to the leaves

        in our case it will print each value and the keys
        that needed to get to it, for example:

        vm0:
            net: lago
            memory: 1024

        will be output as:

        vm0/net/lago
        vm0/memory/1024

            Args:
                info_dict (dict): information to reformat
                delimiter (str): a delimiter for the path components
            Returns:
                str: String representing the formatted info
        """

        def dfs(father, path, acc):
            if isinstance(father, list):
                for child in father:
                    dfs(child, path, acc)
            elif isinstance(father, collections.Mapping):
                for child in sorted(father.items(), key=itemgetter(0)), :
                    dfs(child, path, acc)
            elif isinstance(father, tuple):
                path = copy.copy(path)
                path.append(father[0])
                dfs(father[1], path, acc)
            else:
                # join the last key with it's value
                path[-1] = '{}: {}'.format(path[-1], str(father))
                acc.append(delimiter.join(path))

        result = []
        dfs(info_dict.get('Prefix') or info_dict, [], result)

        return '\n'.join(result)



#
# Copyright 2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#

import os
import guestfs
import logging
import time
import contextlib
from lago.plugins.vm import ExtractPathNoPathError
from lago.utils import LagoException

LOGGER = logging.getLogger(__name__)


class GuestFSError(LagoException):
    pass


@contextlib.contextmanager
def guestfs_conn_ro(disk):
    """
    Open a GuestFS handle and add `disk` in read only mode.

    Args:
        disk(disk path): Path to the disk.

    Yields:
        guestfs.GuestFS: Open GuestFS handle

    Raises:
        :exc:`GuestFSError`: On any guestfs operation failure
    """

    disk_path = os.path.expandvars(disk)
    conn = guestfs.GuestFS(python_return_dict=True)
    conn.add_drive_ro(disk_path)
    conn.set_backend(os.environ.get('LIBGUESTFS_BACKEND', 'direct'))
    try:
        conn.launch()
    except RuntimeError as err:
        LOGGER.debug(err)
        raise GuestFSError(
            'failed starting guestfs in readonly mode for disk: {0}'.
            format(disk)
        )
    try:
        yield conn
    finally:
        conn.shutdown()
        conn.close()


@contextlib.contextmanager
def guestfs_conn_mount_ro(disk_path, disk_root, retries=5, wait=1):
    """
    Open a GuestFS handle with `disk_path` and try mounting the root
    filesystem. `disk_root` is a hint where it should be looked and will
    only be used if GuestFS will not be able to deduce it independently.

    Note that mounting a live guest, can lead to filesystem inconsistencies,
    causing the mount operation to fail. As we use readonly mode, this is
    safe, but the operation itself can still fail. Therefore, this method
    will watch for mount failures and retry 5 times before throwing
    an exception.


    Args:
        disk_path(str): Path to the disk.
        disk_root(str): Hint what is the root device with the OS filesystem.
        retries(int): Number of retries for :func:`~guestfs.GuestFS.mount_ro`
            operation. Note that on each retry a new GuestFS handle will
            be used.
        wait(int): Time to wait between retries.

    Yields:
        guestfs.GuestFS: An open GuestFS handle.

    Raises:
        :exc:`GuestFSError`: On any guestfs operation error, including
            exceeding retries for the :func:`~guestfs.GuestFS.mount_ro`
            operation.

    """

    for attempt in range(retries):
        with guestfs_conn_ro(disk_path) as conn:
            rootfs = find_rootfs(conn, disk_root)
            try:
                conn.mount_ro(rootfs, '/')
            except RuntimeError as err:
                LOGGER.debug(err)
                if attempt < retries - 1:
                    LOGGER.debug(
                        (
                            'failed mounting %s:%s using guestfs, '
                            'attempt %s/%s'
                        ), disk_path, rootfs, attempt + 1, retries
                    )
                    time.sleep(wait)
                    continue
                else:
                    raise GuestFSError(
                        'failed mounting {0}:{1} using guestfs'.format(
                            disk_path, rootfs
                        )
                    )
            yield conn
            try:
                conn.umount(rootfs)
            except RuntimeError as err:
                LOGGER.debug(err)
                raise GuestFSError(
                    ('failed unmounting {0}:{1} using'
                     'guestfs').format(disk_path, rootfs)
                )
            break


def find_rootfs(conn, disk_root):
    """
    Find the image's device root filesystem, and return its path.

    1. Use :func:`guestfs.GuestFS.inspect_os` method. If it returns more than
        one root filesystem or None, try:
    2. Find an exact match of `disk_root` from
        :func:`guestfs.GuestFS.list_filesystems`, if none is found, try:
    3. Return the device that has the substring `disk_root` contained in it,
        from the output of :func:`guestfs.GuestFS.list_filesystems`.

    Args:
        conn(guestfs.GuestFS): Open GuestFS handle.
        disk_root(str): Root device to search for. Note that by default, if
            guestfs can deduce the filesystem, it will not be used.

    Returns:
        str: root device path

    Raises:
        :exc:`GuestFSError` if no root filesystem was found
    """
    rootfs = conn.inspect_os()
    if not rootfs or len(rootfs) > 1:
        filesystems = conn.list_filesystems()
        if disk_root in filesystems:
            rootfs = [disk_root]
        else:
            rootfs = [fs for fs in filesystems.keys() if disk_root in fs]
            if not rootfs:
                raise GuestFSError(
                    'no root fs {0} could be found from list {1}'.format(
                        disk_root, str(filesystems)
                    )
                )
    return sorted(rootfs)[0]


def extract_paths(disk_path, disk_root, paths, ignore_nopath):
    """
    Extract paths from a disk using guestfs

    Args:
        disk_path(str): path to the disk
        disk_root(str): root partition
        paths(list of tuples): files to extract in
            `[(src1, dst1), (src2, dst2)...]` format, if ``srcN`` is a
            directory in the guest, and ``dstN`` does not exist on the host,
            it will be created. If ``srcN`` is a file on the guest, it will be
            copied exactly to ``dstN``
        ignore_nopath(bool): If set to True, ignore paths in the guest that
            do not exit

    Returns:
        None

    Raises:
        :exc:`~lago.plugins.vm.ExtractPathNoPathError`: if a none existing
            path was found on the guest, and `ignore_nopath` is False.
        :exc:`~lago.plugins.vm.ExtractPathError`: on all other failures.
    """

    with guestfs_conn_mount_ro(disk_path, disk_root) as conn:
        for (guest_path, host_path) in paths:
            msg = ('Extracting guestfs://{0} to {1}').format(
                guest_path, host_path
            )

            LOGGER.debug(msg)
            try:
                _copy_path(conn, guest_path, host_path)
            except ExtractPathNoPathError as err:
                if ignore_nopath:
                    LOGGER.debug('%s - ignoring', err)
                else:
                    raise


def _copy_path(conn, guest_path, host_path):
    if conn.is_file(guest_path, followsymlinks=True):
        try:
            conn.download(guest_path, host_path)
        except RuntimeError as err:
            LOGGER.debug(err)
            raise GuestFSError(
                'failed copying file {0} to {1} using guestfs'.format(
                    guest_path, host_path
                )
            )
    elif conn.is_dir(guest_path, followsymlinks=True):
        if not os.path.isdir(host_path):
            os.makedirs(host_path)
        try:
            conn.copy_out(guest_path, host_path)
        except RuntimeError as err:
            LOGGER.debug(err)
            raise GuestFSError(
                'failed copying directory {0} to {1} using guestfs'.format(
                    guest_path, host_path
                )
            )
    else:
        raise ExtractPathNoPathError(
            ('unable to extract {0}: path does not '
             'exist.').format(guest_path)
        )



#
# Copyright 2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#

import logging

from lxml import etree as ET

import lago.providers.libvirt.utils as utils
from lago.utils import LagoException, LagoInitException

LOGGER = logging.getLogger(__name__)


class CPU(object):
    def __init__(self, spec, host_cpu):
        """
        Generate CPU XML node.

        Args:
            spec(dict): VM Lago spec
            host_cpu(lxml.etree.Element): Host cpu capabilities
        """

        self.vcpu_set = False
        # TO-DO: deduce recommended defaults for vcpu_num from host CPU caps
        self.vcpu_num = '2'
        if spec.get('vcpu'):
            self.vcpu_set = True
            self.vcpu_num = spec['vcpu']
        self.cpu_custom = spec.get('cpu_custom')
        self.cpu_model = spec.get('cpu_model')
        self.host_cpu = host_cpu
        self.memory = spec.get('memory')
        self.validate()

        self._cpu_xml = self.generate_cpu_xml()
        self._vcpu_xml = self.generate_vcpu_xml(self.vcpu_num)

    def __iter__(self):
        for node in (self.cpu_xml, self.vcpu_xml):
            yield node

    def validate(self):
        """
        Validate CPU-related VM spec are compatible

        Raises:
            :exc:`~LagoInitException`: if both 'cpu_model' and 'cpu_custom' are
                defined.
        """
        if self.cpu_custom is not None and self.cpu_model:
            raise LagoInitException(
                'Defining both cpu_model and cpu_custom is '
                'not supported.'
            )

    @property
    def cpu_xml(self):
        return self._cpu_xml

    @property
    def vcpu_xml(self):
        return self._vcpu_xml

    @property
    def model(self):
        if self.cpu_model:
            return self.cpu_model
        elif not self.cpu_custom:
            return self.host_cpu.xpath('model')[0].text
        else:
            return self._cpu_xml.xpath('model')[0].text

    @property
    def vendor(self):
        if self.cpu_model:
            return LibvirtCPU.get_cpu_vendor(self.cpu_model)
        elif not self.cpu_custom:
            return self.host_cpu.xpath('vendor')[0].text
        else:
            return LibvirtCPU.get_cpu_vendor(self.model)

    def generate_cpu_xml(self):
        """
        Get CPU XML

        Returns:
            lxml.etree.Element: cpu node
        """
        if self.cpu_custom:
            return self.generate_custom(
                cpu=self.cpu,
                vcpu_num=self.vcpu_num,
                fill_topology=self.vcpu_set
            )
        elif self.cpu_model:
            return self.generate_exact(
                self.cpu_model, vcpu_num=self.vcpu_num, host_cpu=self.host_cpu
            )
        else:
            return self.generate_host_passthrough(self.vcpu_num)

    def generate_vcpu_xml(self, vcpu_num):
        """
        generate_vcpu_xml

        Args:
            vcpu_num(int): number of virtual cpus

        Returns:
            lxml.etree.Element: vcpu XML node
        """

        return self.generate_vcpu(vcpu_num=self.vcpu_num)

    def generate_host_passthrough(self, vcpu_num):
        """
        Generate host-passthrough XML cpu node

        Args:
            vcpu_num(str): number of virtual CPUs

        Returns:
            lxml.etree.Element: CPU XML node
        """

        cpu = ET.Element('cpu', mode='host-passthrough')
        cpu.append(self.generate_topology(vcpu_num))
        if vcpu_num > 1:
            cpu.append(self.generate_numa(vcpu_num))
        return cpu

    def generate_custom(self, cpu, vcpu_num, fill_topology):
        """
        Generate custom CPU model. This method attempts to convert the dict to
        XML, as defined by ``xmltodict.unparse`` method.

        Args:
            cpu(dict): CPU spec
            vcpu_num(int): number of virtual cpus
            fill_topology(bool): if topology is not defined in ``cpu`` and
                ``vcpu`` was not set, will add CPU topology to the generated
                CPU.

        Returns:
            lxml.etree.Element: CPU XML node

        Raises:
            :exc:`~LagoInitException`: when failed to convert dict to XML
        """

        try:
            cpu = utils.dict_to_xml({'cpu': cpu})
        except:
            # TO-DO: print an example here
            raise LagoInitException('conversion of \'cpu\' to XML failed')

        if not cpu.xpath('topology') and fill_topology:
            cpu.append(self.generate_topology(vcpu_num))
        return cpu

    def generate_exact(self, model, vcpu_num, host_cpu):
        """
        Generate exact CPU model with nested virtualization CPU feature.

        Args:
            model(str): libvirt supported CPU model
            vcpu_num(int): number of virtual cpus
            host_cpu(lxml.etree.Element): the host CPU model

        Returns:
            lxml.etree.Element: CPU XML node
        """

        nested = {'Intel': 'vmx', 'AMD': 'svm'}
        cpu = ET.Element('cpu', match='exact')
        ET.SubElement(cpu, 'model').text = model
        cpu.append(self.generate_topology(vcpu_num))

        vendor = host_cpu.findtext('vendor')
        if not nested.get(vendor):
            LOGGER.debug(
                'Unknown vendor: {0}, did not configure nested '
                'virtualization cpu flag on guest.'.format(vendor)
            )
            return cpu

        model_vendor = LibvirtCPU.get_cpu_vendor(family=model)
        if vendor != model_vendor:
            LOGGER.debug(
                (
                    'Not enabling nested virtualization feature, host '
                    'vendor is: {0}, guest vendor: '
                    '{1}'.format(vendor, model_vendor)
                )
            )
            return cpu

        flag = nested[vendor]
        if host_cpu.find('feature/[@name="{0}"]'.format(flag)) is not None:
            cpu.append(self.generate_feature(name=flag))
        else:
            LOGGER.debug(
                (
                    'missing {0} cpu flag on host, nested '
                    'virtualization will probably not '
                    'work.'
                ).format(flag)
            )

        return cpu

    def generate_topology(self, vcpu_num, cores=1, threads=1):
        """
        Generate CPU <topology> XML child

        Args:
            vcpu_num(str): number of virtual CPUs
            cores(int): number of cores
            threads(int): number of threads

        Returns:
            lxml.etree.Element: topology XML element
        """

        return ET.Element(
            'topology',
            sockets=str(vcpu_num),
            cores=str(cores),
            threads=str(threads),
        )

    def generate_numa(self, vcpu_num):
        """
        Generate guest CPU <numa> XML child
        Configures 1, 2 or 4 vCPUs per cell.

        Args:
            vcpu_num(str): number of virtual CPUs

        Returns:
            lxml.etree.Element: numa XML element
        """

        if int(vcpu_num) == 2:
            # 2 vCPUs is a special case.
            # We wish to have 2 cells,
            # with 1 vCPU in each.
            # This is also the common case.
            total_cells = 2
            cpus_per_cell = 1
        elif int(vcpu_num) == 4:
            # 4 vCPU is a special case.
            # We wish to have 2 cells,
            # with 2 vCPUs in each.
            total_cells = 2
            cpus_per_cell = 2
        else:
            cell_info = divmod(int(vcpu_num), 4)
            if cell_info[1] == 0:
                # 4 vCPUs in each cell
                total_cells = cell_info[0]
                cpus_per_cell = 4
            elif cell_info[1] == 2:
                # 2 vCPUs in each cell
                total_cells = (cell_info[0] * 2) + 1
                cpus_per_cell = 2
            else:
                # 1 vCPU per cell...
                total_cells = int(vcpu_num)
                cpus_per_cell = 1

        numa = ET.Element('numa')
        memory_per_cell = divmod(int(self.memory), total_cells)
        LOGGER.debug(
            'numa\n: cpus_per_cell: {0}, total_cells: {1}'.format(
                cpus_per_cell, total_cells
            )
        )
        for cell_id in xrange(0, total_cells):
            first_cpu_in_cell = cell_id * cpus_per_cell
            if cpus_per_cell == 1:
                cpus_in_cell = str(first_cpu_in_cell)
            else:
                cpus_in_cell = '{0}-{1}'.format(
                    first_cpu_in_cell, first_cpu_in_cell + cpus_per_cell - 1
                )
            cell = ET.Element(
                'cell',
                id=str(cell_id),
                cpus=cpus_in_cell,
                memory=str(memory_per_cell[0]),
                unit='MiB',
            )
            numa.append(cell)

        LOGGER.debug('numa:\n{}'.format(ET.tostring(numa, pretty_print=True)))
        return numa

    def generate_vcpu(self, vcpu_num):
        """
        Generate <vcpu> domain XML child

        Args:
            vcpu_num(str): number of virtual cpus

        Returns:
            lxml.etree.Element: vcpu XML element
        """

        vcpu = ET.Element('vcpu')
        vcpu.text = str(vcpu_num)
        return vcpu

    def generate_feature(self, name, policy='require'):
        """
        Generate CPU feature element

        Args:
            name(str): feature name
            policy(str): libvirt feature policy

        Returns:
            lxml.etree.Element: feature XML element
        """

        return ET.Element('feature', policy=policy, name=name)


class LibvirtCPU(object):
    """Query data from /usr/share/libvirt/cpu_map.xml"""

    @classmethod
    def get_cpu_vendor(cls, family, arch='x86'):
        """
        Get CPU vendor, if vendor is not available will return 'generic'

        Args:
            family(str): CPU family
            arch(str): CPU arch

        Returns:
            str: CPU vendor if found otherwise 'generic'
        """

        props = cls.get_cpu_props(family, arch)
        vendor = 'generic'
        try:
            vendor = props.xpath('vendor/@name')[0]
        except IndexError:
            pass
        return vendor

    @classmethod
    def get_cpu_props(cls, family, arch='x86'):
        """
        Get CPU info XML

        Args:
            family(str): CPU family
            arch(str): CPU arch

        Returns:
            lxml.etree.Element: CPU xml

        Raises:
            :exc:`~LagoException`: If no such CPU family exists
        """

        cpus = cls.get_cpus_by_arch(arch)
        try:
            return cpus.xpath('model[@name="{0}"]'.format(family))[0]
        except IndexError:
            raise LagoException('No such CPU family: {0}'.format(family))

    @classmethod
    def get_cpus_by_arch(cls, arch):
        """
        Get all CPUs info by arch

        Args:
            arch(str): CPU architecture

        Returns:
            lxml.etree.element: CPUs by arch XML

        Raises:
            :exc:`~LagoException`: If no such ARCH is found
        """

        with open('/usr/share/libvirt/cpu_map.xml', 'r') as cpu_map:
            cpu_xml = ET.parse(cpu_map)
        try:
            return cpu_xml.xpath('/cpus/arch[@name="{0}"]'.format(arch))[0]
        except IndexError:
            raise LagoException('No such arch: {0}'.format(arch))



#
# Copyright 2015-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
import os

import utils
import logging
import tempfile
from jinja2 import Environment, PackageLoader
import textwrap
import sys

LOGGER = logging.getLogger(__name__)

try:
    import guestfs
except ImportError:
    LOGGER.debug('guestfs not available, ignoring')


def _guestfs_version(default={'major': 1L, 'minor': 20L}):
    if 'guestfs' in sys.modules:
        g = guestfs.GuestFS(python_return_dict=True)
        guestfs_ver = g.version()
        g.close()
    else:
        guestfs_ver = default

    return guestfs_ver


def _render_template(distro, loader, **kwargs):
    env = Environment(
        loader=loader,
        trim_blocks=True,
        lstrip_blocks=True,
    )
    env.filters['dedent'] = textwrap.dedent
    template_name = 'sysprep-{0}.j2'.format(distro)
    template = env.select_template([template_name, 'sysprep-base.j2'])
    sysprep_content = template.render(guestfs_ver=_guestfs_version(), **kwargs)
    with tempfile.NamedTemporaryFile(delete=False) as sysprep_file:
        sysprep_file.write('# {0}\n'.format(template.name))
        sysprep_file.write(sysprep_content)

    LOGGER.debug(
        ('Generated sysprep template '
         'at {0}:\n{1}').format(sysprep_file.name, sysprep_content)
    )
    return sysprep_file.name


def sysprep(disk, distro, loader=None, backend='direct', **kwargs):
    """
    Run virt-sysprep on the ``disk``, commands are built from the distro
    specific template and arguments passed in ``kwargs``. If no template is
    available it will default to ``sysprep-base.j2``.

    Args:
        disk(str): path to disk
        distro(str): distro to render template for
        loader(jinja2.BaseLoader): Jinja2 template loader, if not passed,
            will search Lago's package.
        backend(str): libguestfs backend to use
        **kwargs(dict): environment variables for Jinja2 template

    Returns:
        None

    Raises:
        RuntimeError: On virt-sysprep none 0 exit code.
    """

    if loader is None:
        loader = PackageLoader('lago', 'templates')
    sysprep_file = _render_template(distro, loader=loader, **kwargs)

    cmd = ['virt-sysprep', '-a', disk]
    cmd.extend(['--commands-from-file', sysprep_file])

    env = os.environ.copy()
    if 'LIBGUESTFS_BACKEND' not in env:
        env['LIBGUESTFS_BACKEND'] = backend

    ret = utils.run_command(cmd, env=env)
    if ret:
        raise RuntimeError(
            'Failed to bootstrap %s\ncommand:%s\nstdout:%s\nstderr:%s' % (
                disk,
                ' '.join('"%s"' % elem for elem in cmd),
                ret.out,
                ret.err,
            )
        )



#
# Copyright 2014-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#

from future.utils import raise_from
from future.builtins import super
import functools
import json
import os
import logging
from netaddr import IPNetwork, AddrFormatError
from textwrap import dedent
from lockfile.mkdirlockfile import LockFailed

from .config import config
from . import utils, log_utils

LOGGER = logging.getLogger(__name__)
LogTask = functools.partial(log_utils.LogTask, logger=LOGGER)
LOCK_NAME = 'leases'


class SubnetStore(object):
    """
    SubnetStore object represents a store of subnets used by lago for network
    bridges.

    .. note:: Currently only /24 ranges are handled, and all of them under the
        192.168._min_third_octet to 192.168._max_third_octet ranges.

    The leases are stored under the store's directory (which is specified
    with the `path` argument) as json files with the form::

        [
            "/path/to/prefix/uuid/file",
            "uuid_hash",
        ]

    Where the `uuid_hash` is the 32 char uuid of the prefix (the contents of
    the uuid file at the time of doing the lease).

    The helper class :class:`Lease` is used to abstract the interaction with
    the lease files in the store (each file will be represented with a Lease
    object).

    Cleanup of stale leases is done in a lazy manner during a request for a
    lease. The store will remove at most 1 stale lease in each request (see
    SubnetStore._lease_valid for more info).

    Attributes:
        _path (str): Path to the store, if not specified defaults to the value
            of `lease_dir` in the config
        _cidr (int): Number of bits dedicated for the network address.
            Has a fixed value of 24.
        _subnet_template (str): A template for creating ip address.
            Has a fixed value of `192.168.{}.0`
        _min_third_octet (int): The minimum value of the subnets' last octet.
        _max_third_octet (int): The maximum value of the subnets' last octet.
        _min_subnet (netaddr.IPNetwork): The lowest subnet in the range of
            the store.
        _max_subnet (netaddr.IPNetwork): The highest subnet in the range of
            the store.
    """

    def __init__(
        self,
        path=None,
        min_third_octet=200,
        max_third_octet=255,
    ):

        self._path = path or config['lease_dir']
        self._cidr = 24
        self._subnet_template = '{}/{}'.format('192.168.{}.0', self._cidr)
        self._min_third_octet = min_third_octet
        self._max_third_octet = max_third_octet
        self._min_subnet = IPNetwork(
            self._subnet_template.format(min_third_octet)
        )
        self._max_subnet = IPNetwork(
            self._subnet_template.format(max_third_octet)
        )
        self._validate_lease_dir()

    def _create_lock(self):
        return utils.LockFile(
            path=os.path.join(self.path, LOCK_NAME), timeout=5
        )

    def _validate_lease_dir(self):
        """
        Validate that the directory used by this store exist,
            otherwise create it.
        """
        try:
            if not os.path.isdir(self.path):
                os.makedirs(self.path)
        except OSError as e:
            raise_from(
                LagoSubnetLeaseBadPermissionsException(self.path, e.strerror),
                e
            )

    def acquire(self, uuid_path, subnet=None):
        """
        Lease a free subnet for the given uuid path.
        If subnet is given, try to lease that subnet, otherwise try to lease a
        free subnet.

        Args:
            uuid_path (str): Path to the uuid file of a :class:`lago.Prefix`
            subnet (str): A subnet to lease.
        Returns:
            netaddr.IPAddress: An object which represents the subnet.

        Raises:
            LagoSubnetLeaseException:
                1. If this store is full
                2. If the requested subnet is already taken.
            LagoSubnetLeaseLockException:
                If the lock to self.path can't be acquired.
        """
        try:
            with self._create_lock():
                if subnet:
                    LOGGER.debug('Trying to acquire subnet {}'.format(subnet))
                    acquired_subnet = self._acquire_given_subnet(
                        uuid_path, subnet
                    )
                else:
                    LOGGER.debug('Trying to acquire a free subnet')
                    acquired_subnet = self._acquire(uuid_path)

                return acquired_subnet
        except (utils.TimerException, LockFailed):
            raise LagoSubnetLeaseLockException(self.path)

    def _acquire(self, uuid_path):
        """
        Lease a free network for the given uuid path

        Args:
            uuid_path (str): Path to the uuid file of a :class:`lago.Prefix`

        Returns:
            netaddr.IPNetwork: Which represents the selected subnet

        Raises:
            LagoSubnetLeaseException: If the store is full
        """
        for index in range(self._min_third_octet, self._max_third_octet + 1):
            lease = self.create_lease_object_from_idx(index)
            if self._lease_valid(lease):
                continue
            self._take_lease(lease, uuid_path, safe=False)
            return lease.to_ip_network()

        raise LagoSubnetLeaseStoreFullException(self.get_allowed_range())

    def _acquire_given_subnet(self, uuid_path, subnet):
        """
        Try to create a lease for subnet

        Args:
            uuid_path (str): Path to the uuid file of a :class:`lago.Prefix`
            subnet (str): dotted ipv4 subnet
                (for example ```192.168.200.0```)

        Returns:
            netaddr.IPNetwork: Which represents the selected subnet

        Raises:
            LagoSubnetLeaseException: If the requested subnet is not in the
                range of this store or its already been taken
        """
        lease = self.create_lease_object_from_subnet(subnet)
        self._take_lease(lease, uuid_path)

        return lease.to_ip_network()

    def _lease_valid(self, lease):
        """
        Check if the given lease exist and still has a prefix that owns it.
        If the lease exist but its prefix isn't, remove the lease from this
        store.

        Args:
            lease (lago.subnet_lease.Lease): Object representation of the
                lease

        Returns:
            str or None: If the lease and its prefix exists, return the path
                to the uuid of the prefix, else return None.
        """
        if not lease.exist:
            return None

        if lease.has_env:
            return lease.uuid_path
        else:
            self._release(lease)
            return None

    def _take_lease(self, lease, uuid_path, safe=True):
        """
        Persist the given lease to the store and make the prefix in uuid_path
        his owner

        Args:
            lease(lago.subnet_lease.Lease): Object representation of the lease
            uuid_path (str): Path to the prefix uuid
            safe (bool): If true (the default), validate the the lease
                isn't taken.

        Raises:
            LagoSubnetLeaseException: If safe == True and the lease is already
                taken.
        """
        if safe:
            lease_taken_by = self._lease_valid(lease)
            if lease_taken_by and lease_taken_by != uuid_path:
                raise LagoSubnetLeaseTakenException(
                    lease.subnet, lease_taken_by
                )

        with open(uuid_path) as f:
            uuid = f.read()
        with open(lease.path, 'wt') as f:
            utils.json_dump((uuid_path, uuid), f)

        LOGGER.debug(
            'Assigned subnet lease {} to {}'.format(lease.path, uuid_path)
        )

    def list_leases(self, uuid=None):
        """
        List current subnet leases

        Args:
            uuid(str): Filter the leases by uuid

        Returns:
            list of :class:~Lease: current leases
        """

        try:
            lease_files = os.listdir(self.path)
        except OSError as e:
            raise_from(
                LagoSubnetLeaseBadPermissionsException(self.path, e.strerror),
                e
            )

        leases = [
            self.create_lease_object_from_idx(lease_file.split('.')[0])
            for lease_file in lease_files
        ]
        if not uuid:
            return leases
        else:
            return [lease for lease in leases if lease.uuid == uuid]

    def release(self, subnets):
        """
        Free the lease of the given subnets

        Args:
            subnets (list of str or netaddr.IPAddress): dotted ipv4 subnet in
                CIDR notation (for example ```192.168.200.0/24```) or IPAddress
                object.

        Raises:
            LagoSubnetLeaseException: If subnet is a str and can't be parsed
            LagoSubnetLeaseLockException:
                If the lock to self.path can't be acquired.
        """

        if isinstance(subnets, str) or isinstance(subnets, IPNetwork):
            subnets = [subnets]
        subnets_iter = (
            str(subnet) if isinstance(subnet, IPNetwork) else subnet
            for subnet in subnets
        )
        try:
            with self._create_lock():
                for subnet in subnets_iter:
                    self._release(self.create_lease_object_from_subnet(subnet))
        except (utils.TimerException, LockFailed):
            raise LagoSubnetLeaseLockException(self.path)

    def _release(self, lease):
        """
        Free the given lease

        Args:
            lease (lago.subnet_lease.Lease): The lease to free
        """
        if lease.exist:
            os.unlink(lease.path)
            LOGGER.debug('Removed subnet lease {}'.format(lease.path))

    def _lease_owned(self, lease, current_uuid_path):
        """
        Checks if the given lease is owned by the prefix whose uuid is in
        the given path

        Note:
            The prefix must be also in the same path it was when it took the
            lease

        Args:
            path (str): Path to the lease
            current_uuid_path (str): Path to the uuid to check ownership of

        Returns:
            bool: ``True`` if the given lease in owned by the prefix,
                ``False`` otherwise
        """

        prev_uuid_path, prev_uuid = lease.metadata

        with open(current_uuid_path) as f:
            current_uuid = f.read()

        return \
            current_uuid_path == prev_uuid_path and \
            prev_uuid == current_uuid

    def create_lease_object_from_idx(self, idx):
        """
        Create a lease from self._subnet_template and put idx as its third
        octet.

        Args:
            idx (str): The value of the third octet

        Returns:
            Lease: Lease object which represents the requested subnet.

        Raises:
            LagoSubnetLeaseOutOfRangeException: If the resultant subnet is
            malformed or out of the range of the store.
        """

        return self.create_lease_object_from_subnet(
            self._subnet_template.format(idx)
        )

    def create_lease_object_from_subnet(self, subnet):
        """
        Create a lease from ip in a dotted decimal format,
        (for example `192.168.200.0/24`). the _cidr will be added if not exist
        in `subnet`.

        Args:
            subnet (str): The value of the third octet

        Returns:
            Lease: Lease object which represents the requested subnet.

        Raises:
            LagoSubnetLeaseOutOfRangeException: If the resultant subnet is
            malformed or out of the range of the store.
        """
        if '/' not in subnet:
            subnet = '{}/{}'.format(subnet, self._cidr)

        try:
            if not self.is_leasable_subnet(subnet):
                raise LagoSubnetLeaseOutOfRangeException(
                    subnet, self.get_allowed_range()
                )
        except AddrFormatError:
            raise LagoSubnetLeaseMalformedAddrException(subnet)

        return Lease(store_path=self.path, subnet=subnet)

    def is_leasable_subnet(self, subnet):
        """
        Checks if a given subnet is inside the defined provision-able range

        Args:
            subnet (str): Ip in dotted decimal format with _cidr notation
                (for example `192.168.200.0/24`)

        Returns:
            bool: True if subnet can be parsed into IPNetwork object and is
                inside the range, False otherwise

        Raises:
            netaddr.AddrFormatError: If subnet can not be parsed into an ip.
        """

        return \
            self._min_subnet <= \
            IPNetwork(subnet) <= \
            self._max_subnet

    def get_allowed_range(self):
        """
        Returns:
            str: The range of the store (with lowest and highest subnets as
                the bounds).
        """
        return '{} - {}'.format(self._min_subnet, self._max_subnet)

    @property
    def path(self):
        return self._path


class Lease(object):
    """
    Lease object is an abstraction of a lease file.

    Attributes:
        _store_path (str): Path to the lease's store.
        _subnet (str): The subnet that this lease represents
        _path (str): The path to the lease file
    """

    def __init__(self, store_path, subnet):
        self._store_path = store_path
        self._subnet = subnet
        self._path = None
        self._realise_lease_path()

    def _realise_lease_path(self):
        ip = self.subnet.split('/')[0]
        idx = ip.split('.')[2]
        self._path = os.path.join(self._store_path, '{}.lease'.format(idx))

    def to_ip_network(self):
        return IPNetwork(self.subnet)

    @property
    def valid(self):
        if self.exist:
            return self.has_env
        else:
            return False

    @property
    def metadata(self):
        with open(self.path) as f:
            uuid_path, uuid = json.load(f)

        return uuid_path, uuid

    @property
    def uuid(self):
        return self.metadata[1]

    @property
    def uuid_path(self):
        return self.metadata[0]

    @property
    def has_env(self):
        return self._has_env()

    def _has_env(self, uuid_path=None, uuid=None):
        if not (uuid_path and uuid):
            uuid_path, uuid = self.metadata

        if not os.path.isfile(uuid_path):
            return False

        with open(uuid_path, mode='rt') as f:
            if f.read() == uuid:
                return True
            else:
                return False

    @property
    def exist(self):
        return os.path.isfile(self.path)

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, data):
        self._path = data

    @property
    def subnet(self):
        return self._subnet

    @subnet.setter
    def subnet(self, data):
        self._subnet = data

    def __str__(self):
        return self.subnet


class LagoSubnetLeaseException(utils.LagoException):
    def __init__(self, msg, prv_msg=None):
        if prv_msg is not None:
            msg = msg + '\nOriginal Exception: {0}'.format(prv_msg)
        super().__init__(msg)


class LagoSubnetLeaseLockException(LagoSubnetLeaseException):
    def __init__(self, store_path):
        super().__init__(
            dedent(
                """
                Failed to acquire a lock for store {}.
                This failure can be caused by several reasons:
                1. Another 'lago' environment is using the store.
                2. A stale lock was left in the store.
                3. You don't have R/W permissions to the store.
                """.format(store_path)
            )
        )


class LagoSubnetLeaseStoreFullException(LagoSubnetLeaseException):
    def __init__(self, store_range):
        super().__init__(
            dedent(
                """
                Can't acquire subnet from range {}
                The store of subnets is full.
                You can free subnets by destroying unused lago environments'
                """.format(store_range)
            )
        )


class LagoSubnetLeaseTakenException(LagoSubnetLeaseException):
    def __init__(self, required_subnet, lease_taken_by):
        super().__init__(
            dedent(
                """
                Can't acquire subnet {}.
                The subnet is already taken by {}.
                """.format(required_subnet, lease_taken_by)
            )
        )


class LagoSubnetLeaseOutOfRangeException(LagoSubnetLeaseException):
    def __init__(self, required_subnet, store_range):
        super().__init__(
            dedent(
                """
                Subnet {} is not valid.
                Subnet should be in the range {}.
                """.format(required_subnet, store_range)
            )
        )


class LagoSubnetLeaseMalformedAddrException(LagoSubnetLeaseException):
    def __init__(self, required_subnet):
        super().__init__(
            dedent(
                """
                Address {} is not a valid ip address
                """.format(required_subnet)
            )
        )


class LagoSubnetLeaseBadPermissionsException(LagoSubnetLeaseException):
    def __init__(self, store_path, prv_msg):
        super().__init__(
            dedent(
                """
                    Failed to get access to the store at {}.
                    Please make sure that you have R/W permissions to this
                    directory and that it exists.
                    """.format(store_path)
            ),
            prv_msg=prv_msg,
        )



import functools
import inspect
import wrapt
import logging
from lago.log_utils import get_default_log_formatter


class SDKWrapper(wrapt.ObjectProxy):
    """A proxy object that exposes only methods which were decorated with
    :func:`expose` decorator."""

    def __getattr__(self, name):
        attr = getattr(self.__wrapped__, name)
        return getattr_sdk(attr, name)

    def __dir__(self):
        orig = super(SDKWrapper, self).__dir__()
        filtered = []
        for name in orig:
            attr = getattr(self.__wrapped__, name)
            try:
                getattr_sdk(attr, name)
                filtered.append(name)
            except AttributeError:
                pass
        return filtered


class SDKMethod(object):
    """Metadata to store inside the decorated function"""

    def __init__(self, name):
        self.name = name


def expose(func):
    """
    Decorator to be used with :class:`SDKWrapper`. This decorator indicates
    that the wrapped method or class should be exposed in the proxied object.

    Args:
        func(types.FunctionType/types.MethodType): function to decorate

    Returns:
        None
    """

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        return func(*args, **kwargs)

    if inspect.isclass(func):
        wrapped._sdkmetaclass = SDKMethod(func.__name__)
    else:
        wrapped._sdkmeta = SDKMethod(func.__name__)
    return wrapped


def getattr_sdk(attr, name):
    """
    Filter SDK attributes

    Args:
        attr(attribute): Attribute as returned by :func:`getattr`.
        name(str): Attribute name.

    Returns:
        `attr` if passed.
    """
    if inspect.isroutine(attr):
        if hasattr(attr, '_sdkmeta'):
            return attr
    raise AttributeError(name)


def setup_sdk_logging(logfile=None, loglevel=logging.INFO):
    """
    Setup a NullHandler to the root logger. If ``logfile`` is passed,
    additionally add a FileHandler in ``loglevel`` level.

    Args:
        logfile(str): A path to setup a log file.
        loglevel(int): :mod:`logging` log level.

    Returns:
        None
    """

    logging.root.setLevel(logging.DEBUG)
    logging.root.addHandler(logging.NullHandler())
    if logfile:
        fh = logging.FileHandler(logfile)
        fh.setLevel(loglevel)
        fh.setFormatter(get_default_log_formatter())
        logging.root.addHandler(fh)



#
# Copyright 2016-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#

from lago.plugins import vm
from lago import ssh


class DefaultVM(vm.VMPlugin):
    pass


class SSHVMProvider(vm.VMProviderPlugin):
    def start(self, *args, **kwargs):
        pass

    def stop(self, *args, **kwargs):
        pass

    def defined(self, *args, **kwargs):
        return True

    def bootstrap(self, *args, **kwargs):
        pass

    def state(self, *args, **kwargs):
        return 'running'

    def running(self, *args, **kwargs):
        try:
            ssh.get_ssh_client(
                ip_addr=self.vm.ip(),
                host_name=self.vm.name(),
                propagate_fail=False,
                ssh_tries=1,
                ssh_key=self.vm.virt_env.prefix.paths.ssh_id_rsa(),
                username=self.vm._spec.get('ssh-user'),
                password=self.vm._spec.get('ssh-password'),
            )
        except ssh.LagoSSHTimeoutException:
            return False

        return True

    def create_snapshot(self, name, *args, **kwargs):
        pass

    def revert_snapshot(self, name, *args, **kwargs):
        pass
